"""Utilities for parsing typeshed files."""

import collections
import os
import re
from typing import Sequence

from pytype import module_utils
from pytype import pytype_source_utils
from pytype import utils
from pytype.pyi import parser
from pytype.pytd import builtin_stubs

import toml


def _get_module_names_in_path(lister, path, python_version):
  """Get module names for all .pyi files in the given path."""
  names = set()
  try:
    contents = list(lister(path, python_version))
  except pytype_source_utils.NoSuchDirectory:
    pass
  else:
    for filename in contents:
      names.add(module_utils.path_to_module_name(filename))
  return names


class Typeshed:
  """A typeshed installation.

  The location is either retrieved from the environment variable
  "TYPESHED_HOME" (if set) or otherwise assumed to be directly under
  pytype (i.e., /{some_path}/pytype/typeshed).
  """

  # Text file of typeshed entries that will not be loaded.
  # The path is relative to typeshed's root directory, e.g. if you set this to
  # "missing.txt" you need to create $TYPESHED_HOME/missing.txt or
  # pytype/typeshed/missing.txt
  # For testing, this file must contain the entry 'stdlib/3/pytypecanary'.
  MISSING_FILE = None

  def __init__(self):
    self._env_home = home = os.getenv("TYPESHED_HOME")
    if home:
      if not os.path.isdir(home):
        raise OSError("Could not find a typeshed installation in "
                      "$TYPESHED_HOME directory %s" % home)
      self._root = home
    else:
      self._root = pytype_source_utils.get_full_path("typeshed")
    self._missing = frozenset(self._load_missing())
    self._stdlib_versions = self._load_stdlib_versions()
    self._third_party_packages = self._load_third_party_packages()

  def _load_file(self, path):
    if self._env_home:
      filename = os.path.join(self._env_home, path)
      with open(filename) as f:
        return filename, f.read()
    else:
      filepath = os.path.join(self._root, path)
      return filepath, pytype_source_utils.load_text_file(filepath)

  def _file_exists(self, relpath):
    """Checks whether the given path, relative to the typeshed root, exists."""
    if self._env_home:
      return os.path.exists(os.path.join(self._root, relpath))
    try:
      # For a non-par pytype installation, load_text_file will either succeed,
      # raise FileNotFoundError, or raise IsADirectoryError.
      # For a par installation, load_text_file will raise FileNotFoundError for
      # both a nonexistent file and a directory.
      pytype_source_utils.load_text_file(os.path.join("typeshed", relpath))
    except FileNotFoundError:
      try:
        # For a non-par installation, we know at this point that relpath does
        # not exist, so _list_files will always raise NoSuchDirectory. For a par
        # installation, we use _list_files to check whether the directory
        # exists; a non-existent directory will produce an empty generator.
        next(self._list_files(relpath))
      except (pytype_source_utils.NoSuchDirectory, StopIteration):
        return False
    except IsADirectoryError:
      return True
    return True

  def _list_files(self, basedir):
    """Lists files recursively in a basedir relative to typeshed root."""
    if self._env_home:
      fs = pytype_source_utils.list_files(os.path.join(self._root, basedir))
    else:
      fs = pytype_source_utils.list_pytype_files(
          os.path.join("typeshed", basedir))
    return [f for f in fs if "@python2" not in f]

  def _load_missing(self):
    if not self.MISSING_FILE:
      return set()
    _, text = self._load_file(self.MISSING_FILE)
    return {line.strip() for line in text.split("\n")
            if line and "@python2" not in line}

  def _load_stdlib_versions(self):
    """Loads the contents of typeshed/stdlib/VERSIONS.

    VERSIONS lists the stdlib modules with the Python version in which they were
    first added, in the format `{module}: {min_major}.{min_minor}-` or
    `{module}: {min_major}.{min_minor}-{max_major}.{max_minor}`.

    Returns:
      A mapping from module name to version range in the format
        {name: ((min_major, min_minor), (max_major, max_minor))}
      The max tuple can be `None`.
    """
    _, text = self._load_file(os.path.join("stdlib", "VERSIONS"))
    versions = {}
    for line in text.splitlines():
      line2 = line.split("#")[0].strip()
      if not line2:
        continue
      match = re.fullmatch(r"(.+): (\d)\.(\d+)(?:-(?:(\d)\.(\d+))?)?", line2)
      assert match
      module, min_major, min_minor, max_major, max_minor = match.groups()
      minimum = (int(min_major), int(min_minor))
      maximum = ((int(max_major), int(max_minor))
                 if max_major is not None and max_minor is not None
                 else None)
      versions[module] = minimum, maximum
    return versions

  def _load_third_party_packages(self):
    """Loads package and Python version information for typeshed/stubs/.

    stubs/ contains type information for third-party packages. Each top-level
    directory corresponds to one PyPI package and contains one or more modules,
    plus a metadata file (METADATA.toml). If a package supports Python 2,
    it will either have separate stubs in a @python2 subdirectory or
    the `python2 = true` flag is set in METADATA.toml. If a package supports
    Python 3, it will have at least one module or package on the top level
    and the `python3 = false` flag is not set in METADATA.toml. Finally, the
    top-level directory may contain a @tests subdirectory for typeshed testing.

    Returns:
      A mapping from module name to a set of package names.
    """
    modules = collections.defaultdict(set)
    top_level_stubs = set()  # packages with stub files outside @python2
    no_py3_meta = set()  # packages with `python3 = false` metadata entry
    for third_party_file in self._list_files("stubs"):
      parts = third_party_file.split(os.path.sep)
      if parts[-1] == "METADATA.toml":  # {package}/METADATA.toml
        _, md_file = self._load_file(os.path.join("stubs", third_party_file))
        metadata = toml.loads(md_file)
        if not metadata.get("python3", True):
          no_py3_meta.add(parts[0])
      elif parts[1] != "@tests":  # {package}/{module}[/{submodule}]
        if parts[-1].endswith(".pyi"):
          top_level_stubs.add(parts[0])
        name, _ = os.path.splitext(parts[1])
        modules[parts[0]].add(name)
    py3_stubs = top_level_stubs - no_py3_meta
    packages = collections.defaultdict(set)
    for package, names in modules.items():
      for name in names:
        if package in py3_stubs:
          packages[name].add(package)
    return packages

  @property
  def missing(self):
    """Set of known-missing typeshed modules, as strings of paths."""
    return self._missing

  @property
  def root(self):
    """Path of typeshed's root directory.

    Returns:
      Base of filenames returned by get_module_file(). Not guaranteed to exist
      if typeshed is bundled with pytype.
    """
    return self._root

  def get_module_file(self, toplevel, module, version):
    """Get the contents of a typeshed .pyi file.

    Arguments:
      toplevel: the top-level directory within typeshed/, "builtins", "stdlib",
        or "third_party". "builtins" doesn't exist but is requested because
        there exists a pytype pyi directory with this name, and "third_party"
        corresponds to the the typeshed/stubs/ directory.
      module: module name (e.g., "sys" or "__builtins__"). Can contain dots, if
        it's a submodule. Package names should omit the "__init__" suffix (e.g.,
        pass in "os", not "os.__init__").
      version: The Python version. (major, minor)

    Returns:
      A tuple with the filename and contents of the file
    Raises:
      IOError: if file not found
    """
    module_parts = module.split(".")
    module_path = os.path.join(*module_parts)
    paths = []
    if toplevel == "stdlib":
      # Stubs for the stdlib 'foo' module are located in stdlib/foo.
      # The VERSIONS file tells us whether stdlib/foo exists and what versions
      # it targets.
      path = os.path.join(toplevel, module_path)
      if (self._is_module_in_typeshed(module_parts, version) or
          path in self.missing):
        paths.append(path)
    elif toplevel == "third_party":
      # For third-party modules, we grab the alphabetically first package that
      # provides a module with the specified name in the right version.
      # TODO(rechen): It would be more correct to check what packages are
      # currently installed and only consider those.
      for package in sorted(self._third_party_packages[module_parts[0]]):
        paths.append(os.path.join("stubs", package, module_path))
    for path_rel in paths:
      # Give precedence to MISSING_FILE
      if path_rel in self.missing:
        return (os.path.join(self._root, "nonexistent", path_rel + ".pyi"),
                builtin_stubs.DEFAULT_SRC)
      for path in [os.path.join(path_rel, "__init__.pyi"), path_rel + ".pyi"]:
        try:
          name, src = self._load_file(path)
          return name, src
        except OSError:
          pass
    raise OSError(f"Couldn't find {module}")

  def _lookup_stdlib_version(self, module_parts: Sequence[str]):
    """Looks up the prefix chain until we find the module in stdlib/VERSIONS."""
    index = len(module_parts)
    while index > 0:
      name = ".".join(module_parts[:index])
      if name in self._stdlib_versions:
        return self._stdlib_versions[name]
      index -= 1
    return None

  def _is_module_in_typeshed(self, module_parts, version):
    assert module_parts[-1] != "__init__", module_parts
    version_info = self._lookup_stdlib_version(module_parts)
    if version_info is None:
      return False
    min_version, max_version = version_info
    return (min_version <= version and
            (max_version is None or max_version >= version))

  def get_typeshed_paths(self):
    """Gets the paths to typeshed's version-specific pyi files."""
    typeshed_subdirs = ["stdlib"]
    for packages in self._third_party_packages.values():
      for package in packages:
        typeshed_subdirs.append(os.path.join("stubs", package))
    return [os.path.join(self._root, d) for d in typeshed_subdirs]

  def get_pytd_paths(self):
    """Gets the paths to pytype's version-specific pytd files."""
    return [pytype_source_utils.get_full_path(d)
            for d in ("stubs/builtins", "stubs/stdlib")]

  def _list_modules(self, path, python_version):
    """Lists modules for _get_module_names_in_path."""
    for filename in self._list_files(path):
      if filename in ("VERSIONS", "METADATA.toml"):
        # stdlib/VERSIONS, stubs/{package}/METADATA.toml are metadata files.
        continue
      parts = path.split("/")
      if "stdlib" in parts:
        # Check supported versions for stubs directly in stdlib/.
        module_parts = module_utils.strip_init_suffix(
            os.path.splitext(filename)[0].split("/"))
        if not self._is_module_in_typeshed(module_parts, python_version):
          continue
      yield filename

  def _get_missing_modules(self):
    """Gets module names from the `missing` list."""
    module_names = set()
    for f in self.missing:
      parts = f.split("/")
      if parts[0] == "stdlib":
        start_index = 1  # remove stdlib/ prefix
      else:
        assert parts[0] == "stubs"
        start_index = 2  # remove stubs/{package}/ prefix
      filename = "/".join(parts[start_index:])
      module_names.add(filename.replace("/", "."))
    return module_names

  def get_all_module_names(self, python_version):
    """Get the names of all modules in typeshed or bundled with pytype."""
    module_names = set()
    for abspath in self.get_typeshed_paths():
      relpath = abspath.rpartition("typeshed/")[-1]
      module_names |= _get_module_names_in_path(
          self._list_modules, relpath, python_version)
    for abspath in self.get_pytd_paths():
      relpath = abspath.rpartition("pytype/")[-1]
      module_names |= _get_module_names_in_path(
          lambda path, _: pytype_source_utils.list_pytype_files(path),
          relpath, python_version)
    # Also load modules not in typeshed, so that we have a dummy entry for them.
    module_names |= self._get_missing_modules()
    assert "ctypes" in module_names  # sanity check
    return module_names

  def read_blacklist(self):
    """Read the typeshed blacklist."""
    _, text = self._load_file(os.path.join("tests", "pytype_exclude_list.txt"))
    for line in text.splitlines():
      if "#" in line:
        line = line[:line.index("#")]
      line = line.strip()
      if line and "@python2" not in line:
        yield line

  def blacklisted_modules(self):
    """Return the blacklist, as a list of module names. E.g. ["x", "y.z"]."""
    for path in self.read_blacklist():
      parts = path.split(os.path.sep)  # E.g. ["stdlib", "html", "parser.pyi"]
      if parts[0] == "stdlib":
        filename = os.path.sep.join(parts[1:])
      else:
        filename = os.path.sep.join(parts[2:])
      mod = module_utils.path_to_module_name(filename)
      if mod:
        yield mod


_typeshed = None


def _get_typeshed():
  """Get the global Typeshed instance."""
  global _typeshed
  if _typeshed is None:
    try:
      _typeshed = Typeshed()
    except OSError as e:
      # This happens if typeshed is not available. Which is a setup error
      # and should be propagated to the user. The IOError is caught further up
      # in the stack.
      raise utils.UsageError(f"Couldn't initialize typeshed:\n {str(e)}")
  return _typeshed


def parse_type_definition(pyi_subdir, module, options):
  """Load and parse a *.pyi from typeshed.

  Args:
    pyi_subdir: the directory where the module should be found.
    module: the module name (without any file extension or "__init__" suffix).
    options: the parsing options.

  Returns:
    None if the module doesn't have a definition.
    Else a tuple of the filename and the AST of the module.
  """
  typeshed = _get_typeshed()
  try:
    filename, src = typeshed.get_module_file(
        pyi_subdir, module, options.python_version)
  except OSError:
    return None

  ast = parser.parse_string(src, filename=filename, name=module,
                            options=options)
  return filename, ast



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable as Serialisable

class ChartsheetProperties(Serialisable):
    tagname: str
    published: Any
    codeName: Any
    tabColor: Any
    __elements__: Any
    def __init__(self, published: Any | None = ..., codeName: Any | None = ..., tabColor: Any | None = ...) -> None: ...



"""Code related to sequence merging and MRO."""

from pytype.pytd import pytd


def MergeSequences(seqs):
  """Merge a sequence of sequences into a single sequence.

  This code is copied from https://www.python.org/download/releases/2.3/mro/
  with print statements removed and modified to take a sequence of sequences.
  We use it to merge both MROs and class templates.

  Args:
    seqs: A sequence of sequences.

  Returns:
    A single sequence in which every element of the input sequences appears
    exactly once and local precedence order is preserved.

  Raises:
    ValueError: If the merge is impossible.
  """
  res = []
  while True:
    if not any(seqs):  # any empty subsequence left?
      return res
    for seq in seqs:  # find merge candidates among seq heads
      if not seq:
        continue
      cand = seq[0]
      if getattr(cand, "SINGLETON", False):
        # Special class. Cycles are allowed. Emit and remove duplicates.
        seqs = [[s for s in seq if s != cand] for seq in seqs]  # pylint: disable=g-complex-comprehension
        break
      if any(s for s in seqs if cand in s[1:] and s is not seq):
        cand = None  # reject candidate
      else:
        # Remove and emit. The candidate can be head of more than one list.
        for other_seq in seqs:
          if other_seq and other_seq[0] == cand:
            del other_seq[0]
        break
    if cand is None:
      raise ValueError
    res.append(cand)


def Dedup(seq):
  """Return a sequence in the same order, but with duplicates removed."""
  seen = set()
  result = []
  for s in seq:
    if s not in seen:
      result.append(s)
    seen.add(s)
  return result


class MROError(Exception):  # pylint: disable=g-bad-exception-name

  def __init__(self, seqs):
    super().__init__()
    self.mro_seqs = seqs


def MROMerge(input_seqs):
  """Merge a sequence of MROs into a single resulting MRO.

  Args:
    input_seqs: A sequence of MRO sequences.

  Returns:
    A single resulting MRO.

  Raises:
    MROError: If we discovered an illegal inheritance.
  """
  seqs = [Dedup(s) for s in input_seqs]
  try:
    return MergeSequences(seqs)
  except ValueError as e:
    raise MROError(input_seqs) from e


def _GetClass(t, lookup_ast):
  if t.cls:
    return t.cls
  if lookup_ast:
    return lookup_ast.Lookup(t.name)
  raise AttributeError(f"Class not found: {t.name}")


def _Degenerify(types):
  return [t.base_type if isinstance(t, pytd.GenericType) else t for t in types]


def _ComputeMRO(t, mros, lookup_ast):
  """Compute the MRO."""
  if isinstance(t, pytd.ClassType):
    if t not in mros:
      mros[t] = None
      base_mros = []
      for base in _GetClass(t, lookup_ast).bases:
        if base in mros:
          if mros[base] is None:
            raise MROError([[t]])
          else:
            base_mro = mros[base]
        else:
          base_mro = _ComputeMRO(base, mros, lookup_ast)
        base_mros.append(base_mro)
      mros[t] = tuple(
          MROMerge([[t]] + base_mros + [_Degenerify(
              _GetClass(t, lookup_ast).bases)]))
    return mros[t]
  elif isinstance(t, pytd.GenericType):
    return _ComputeMRO(t.base_type, mros, lookup_ast)
  else:
    return [t]


def GetBasesInMRO(cls, lookup_ast=None):
  """Get the given class's bases in Python's method resolution order."""
  mros = {}
  base_mros = []
  for p in cls.bases:
    base_mros.append(_ComputeMRO(p, mros, lookup_ast))
  return tuple(MROMerge(base_mros + [_Degenerify(cls.bases)]))



import sys
from typing import Callable
from xml.etree.ElementTree import Element

XINCLUDE: str
XINCLUDE_INCLUDE: str
XINCLUDE_FALLBACK: str

if sys.version_info >= (3, 9):
    DEFAULT_MAX_INCLUSION_DEPTH: int

class FatalIncludeError(SyntaxError): ...

def default_loader(href: str | bytes | int, parse: str, encoding: str | None = ...) -> str | Element: ...

# TODO: loader is of type default_loader ie it takes a callable that has the
# same signature as default_loader. But default_loader has a keyword argument
# Which can't be represented using Callable...
if sys.version_info >= (3, 9):
    def include(
        elem: Element, loader: Callable[..., str | Element] | None = ..., base_url: str | None = ..., max_depth: int | None = ...
    ) -> None: ...

    class LimitedRecursiveIncludeError(FatalIncludeError): ...

else:
    def include(elem: Element, loader: Callable[..., str | Element] | None = ...) -> None: ...



from typing import Text

class NetrcParseError(Exception):
    filename: str | None
    lineno: int | None
    msg: str
    def __init__(self, msg: str, filename: Text | None = ..., lineno: int | None = ...) -> None: ...

# (login, account, password) tuple
_NetrcTuple = tuple[str, str | None, str | None]

class netrc:
    hosts: dict[str, _NetrcTuple]
    macros: dict[str, list[str]]
    def __init__(self, file: Text | None = ...) -> None: ...
    def authenticators(self, host: str) -> _NetrcTuple | None: ...



"""Trace function arguments, return values and calls to other functions."""

import dataclasses

from typing import Any, List

from pytype.pytd import escape
from pytype.pytd import pytd
from pytype.pytd import pytd_utils


@dataclasses.dataclass
class Attr:
  name: str
  node_type: str
  type: Any
  attrib: str
  location: str


@dataclasses.dataclass
class Arg:
  name: str
  node_type: str
  type: Any


@dataclasses.dataclass
class Param:
  name: str
  type: Any


@dataclasses.dataclass
class Call:
  function_id: str
  args: List[Arg]
  location: str


@dataclasses.dataclass
class Function:
  id: str
  params: List[Any] = dataclasses.field(default_factory=list)
  param_attrs: List[Any] = dataclasses.field(default_factory=list)
  local_attrs: List[Any] = dataclasses.field(default_factory=list)
  calls: List[Any] = dataclasses.field(default_factory=list)
  ret: Any = dataclasses.field(default=None)
  location: Any = dataclasses.field(default=None)


def unknown_to_any(typename):
  if escape.UNKNOWN in typename:
    return 'typing.Any'
  return typename


def unwrap_type(typ):
  if isinstance(typ, (pytd.ClassType, pytd.NamedType)):
    typ_name = typ.name
  elif isinstance(typ, pytd.UnionType):
    typ_name = 'Union[' + ', '.join(unwrap_type(t) for t in typ.type_list) + ']'
  elif isinstance(typ, pytd.AnythingType):
    typ_name = 'typing.Any'
  else:
    typ_name = pytd_utils.Print(typ)
  return unknown_to_any(typ_name)


def get_function_params(pytd_fn):
  """Collect function param types from pytype."""
  # We have turned call records on in the indexer, so a function will have a
  # pytd signature for every tuple of call args. Here we iterate through those
  # signatures and set every param's type to the union of every non-"unknown"
  # call type for that param.
  params = {}
  for sig in pytd_fn.signatures:
    for p in sig.params:
      if p.name not in params:
        params[p.name] = []
      if escape.UNKNOWN not in str(p.type):
        params[p.name].append(p.type)
  for k in params:
    params[k] = pytd_utils.JoinTypes(params[k])
  return [(k, unwrap_type(v)) for k, v in params.items()]


class FunctionMap:
  """Collect a map of function types and outbound callgraph edges."""

  def __init__(self, index):
    self.index = index
    self.fmap = self.init_from_index(index)

  def pytd_of_fn(self, f):
    """Get the pytype pytd function definition."""
    if f.data and f.data[0]:
      d = f.data[0][0]
      try:
        return self.index.get_pytd_def(d, f.name)
      except:  # pylint: disable=bare-except
        # We sometimes get Instance(PyTDClass(str)) here, which throws an
        # exception in get_pytd_def, possibly due to other earlier problems.
        # Don't crash the indexer if this happens.
        return None
    else:
      # TODO(mdemello): log this
      return None

  def init_from_index(self, index):
    """Initialize the function map."""
    out = {}
    fn_defs = [(k, v) for k, v in index.defs.items() if v.typ == 'FunctionDef']
    for fn_id, fn in fn_defs:
      pytd_fn = self.pytd_of_fn(fn)
      if isinstance(pytd_fn, pytd.Function):
        params = get_function_params(pytd_fn)
      else:
        # Sometimes pytype cannot infer the type of a function, and falls back
        # to Any. Don't crash the indexer if this happens.
        params = []
      params = [Param(name, typ) for name, typ in params]
      ret = index.envs[fn_id].ret
      if fn_id in index.locs:
        location = index.locs[fn_id][-1].location
      else:
        location = None
      out[fn_id] = Function(
          id=fn_id, params=params, ret=ret, location=location)
    # Add a slot for "module" to record function calls made at top-level
    out['module'] = Function(id='module')
    return out

  def add_attr(self, ref, defn):
    """Add an attr access within a function body."""
    attrib = ref.name
    scope = ref.ref_scope
    if scope not in self.fmap:
      # This call was not within a function body.
      return

    try:
      d = self.index.envs[scope].env[ref.target]
    except KeyError:
      return

    typename = unknown_to_any(defn.typename)
    attr_access = Attr(
        name=d.name,
        node_type=d.typ,
        type=typename,
        attrib=attrib,
        location=ref.location)
    fn = self.fmap[scope]
    if attr_access.node_type == 'Param':
      fn.param_attrs.append(attr_access)
    else:
      fn.local_attrs.append(attr_access)

  def add_param_def(self, ref, defn):
    """Add a function parameter definition."""
    fn = self.fmap[ref.ref_scope]
    for param in fn.params:
      # Don't override a type inferred from call sites.
      if param.name == defn.name and param.type in ('nothing', 'typing.Any'):
        param.type = unwrap_type(self.index.get_pytd(ref.data[0]))
        break

  def add_link(self, ref, defn):
    if ref.typ == 'Attribute':
      self.add_attr(ref, defn)
    if defn.typ == 'Param':
      self.add_param_def(ref, defn)

  def add_call(self, call):
    """Add a function call."""
    scope = call.scope
    if scope not in self.fmap:
      # This call was not within a function body.
      return
    env = self.index.envs[scope]
    args = []
    for name in call.args:
      if name in env.env:
        defn = env.env[name]
        node_type = defn.typ
        typename = unknown_to_any(defn.typename)
      else:
        node_type = None
        typename = 'typing.Any'
      args.append(Arg(name, node_type, typename))
    self.fmap[scope].calls.append(
        Call(call.func, args, call.location))


def collect_function_map(index):
  """Track types and outgoing calls within a function."""

  fns = FunctionMap(index)

  # Collect methods and attribute accesses
  for ref, defn in index.links:
    fns.add_link(ref, defn)

  # Collect function calls
  for call in index.calls:
    fns.add_call(call)

  return fns.fmap



"""Too-complex program."""


v = [0, 0.1, 1j, "foo", slice(0, 3), "foo", None, type, max]
w = [0, 0.1, 1j, "foo", slice(0, 3), "foo", None, type, max]
x = [0, 0.1, 1j, "foo", slice(0, 3), "foo", None, type, max]
y = [0, 0.1, 1j, "foo", slice(0, 3), "foo", None, type, max]
z = zip(v, w, x, y)



from .sql.sqltypes import (
    ARRAY as ARRAY,
    BIGINT as BIGINT,
    BINARY as BINARY,
    BLOB as BLOB,
    BOOLEAN as BOOLEAN,
    CHAR as CHAR,
    CLOB as CLOB,
    DATE as DATE,
    DATETIME as DATETIME,
    DECIMAL as DECIMAL,
    FLOAT as FLOAT,
    INT as INT,
    INTEGER as INTEGER,
    JSON as JSON,
    NCHAR as NCHAR,
    NUMERIC as NUMERIC,
    NVARCHAR as NVARCHAR,
    REAL as REAL,
    SMALLINT as SMALLINT,
    TEXT as TEXT,
    TIME as TIME,
    TIMESTAMP as TIMESTAMP,
    VARBINARY as VARBINARY,
    VARCHAR as VARCHAR,
    BigInteger as BigInteger,
    Boolean as Boolean,
    Concatenable as Concatenable,
    Date as Date,
    DateTime as DateTime,
    Enum as Enum,
    Float as Float,
    Indexable as Indexable,
    Integer as Integer,
    Interval as Interval,
    LargeBinary as LargeBinary,
    MatchType as MatchType,
    NullType as NullType,
    Numeric as Numeric,
    PickleType as PickleType,
    SmallInteger as SmallInteger,
    String as String,
    Text as Text,
    Time as Time,
    TupleType as TupleType,
    Unicode as Unicode,
    UnicodeText as UnicodeText,
    _Binary as _Binary,
)
from .sql.type_api import (
    ExternalType as ExternalType,
    TypeDecorator as TypeDecorator,
    TypeEngine as TypeEngine,
    UserDefinedType as UserDefinedType,
)

__all__ = [
    "TypeEngine",
    "TypeDecorator",
    "UserDefinedType",
    "ExternalType",
    "INT",
    "CHAR",
    "VARCHAR",
    "NCHAR",
    "NVARCHAR",
    "TEXT",
    "Text",
    "FLOAT",
    "NUMERIC",
    "REAL",
    "DECIMAL",
    "TIMESTAMP",
    "DATETIME",
    "CLOB",
    "BLOB",
    "BINARY",
    "VARBINARY",
    "BOOLEAN",
    "BIGINT",
    "SMALLINT",
    "INTEGER",
    "DATE",
    "TIME",
    "TupleType",
    "String",
    "Integer",
    "SmallInteger",
    "BigInteger",
    "Numeric",
    "Float",
    "DateTime",
    "Date",
    "Time",
    "LargeBinary",
    "Boolean",
    "Unicode",
    "Concatenable",
    "UnicodeText",
    "PickleType",
    "Interval",
    "Enum",
    "Indexable",
    "ARRAY",
    "JSON",
]



from typing import Any

def prepare_headers(oauth_params, headers: Any | None = ..., realm: Any | None = ...): ...
def prepare_form_encoded_body(oauth_params, body): ...
def prepare_request_uri_query(oauth_params, uri): ...



from typing import Any as _Any

import sqlalchemy.types as sqltypes

from ...sql import expression

def Any(other, arrexpr, operator=...): ...
def All(other, arrexpr, operator=...): ...

class array(expression.ClauseList, expression.ColumnElement[_Any]):
    __visit_name__: str
    stringify_dialect: str
    inherit_cache: bool
    type: _Any
    def __init__(self, clauses, **kw) -> None: ...
    def self_group(self, against: _Any | None = ...): ...

CONTAINS: _Any
CONTAINED_BY: _Any
OVERLAP: _Any

class ARRAY(sqltypes.ARRAY):
    class Comparator(sqltypes.ARRAY.Comparator):
        def contains(self, other, **kwargs): ...
        def contained_by(self, other): ...
        def overlap(self, other): ...
    comparator_factory: _Any
    item_type: _Any
    as_tuple: _Any
    dimensions: _Any
    zero_indexes: _Any
    def __init__(self, item_type, as_tuple: bool = ..., dimensions: _Any | None = ..., zero_indexes: bool = ...) -> None: ...
    @property
    def hashable(self): ...
    @property
    def python_type(self): ...
    def compare_values(self, x, y): ...
    def bind_expression(self, bindvalue): ...
    def bind_processor(self, dialect): ...
    def result_processor(self, dialect, coltype): ...



import sys
from _typeshed import Self, SupportsRead, SupportsReadline
from socket import socket
from ssl import SSLContext
from types import TracebackType
from typing import Any, Callable, Iterable, Iterator, TextIO
from typing_extensions import Literal

__all__ = ["FTP", "error_reply", "error_temp", "error_perm", "error_proto", "all_errors", "FTP_TLS"]

MSG_OOB: Literal[1]
FTP_PORT: Literal[21]
MAXLINE: Literal[8192]
CRLF: Literal["\r\n"]
B_CRLF: Literal[b"\r\n"]

class Error(Exception): ...
class error_reply(Error): ...
class error_temp(Error): ...
class error_perm(Error): ...
class error_proto(Error): ...

all_errors: tuple[type[Exception], ...]

class FTP:
    debugging: int
    host: str
    port: int
    maxline: int
    sock: socket | None
    welcome: str | None
    passiveserver: int
    timeout: int
    af: int
    lastresp: str
    file: TextIO | None
    encoding: str
    def __enter__(self: Self) -> Self: ...
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None: ...
    source_address: tuple[str, int] | None
    if sys.version_info >= (3, 9):
        def __init__(
            self,
            host: str = ...,
            user: str = ...,
            passwd: str = ...,
            acct: str = ...,
            timeout: float = ...,
            source_address: tuple[str, int] | None = ...,
            *,
            encoding: str = ...,
        ) -> None: ...
    else:
        def __init__(
            self,
            host: str = ...,
            user: str = ...,
            passwd: str = ...,
            acct: str = ...,
            timeout: float = ...,
            source_address: tuple[str, int] | None = ...,
        ) -> None: ...

    def connect(
        self, host: str = ..., port: int = ..., timeout: float = ..., source_address: tuple[str, int] | None = ...
    ) -> str: ...
    def getwelcome(self) -> str: ...
    def set_debuglevel(self, level: int) -> None: ...
    def debug(self, level: int) -> None: ...
    def set_pasv(self, val: bool | int) -> None: ...
    def sanitize(self, s: str) -> str: ...
    def putline(self, line: str) -> None: ...
    def putcmd(self, line: str) -> None: ...
    def getline(self) -> str: ...
    def getmultiline(self) -> str: ...
    def getresp(self) -> str: ...
    def voidresp(self) -> str: ...
    def abort(self) -> str: ...
    def sendcmd(self, cmd: str) -> str: ...
    def voidcmd(self, cmd: str) -> str: ...
    def sendport(self, host: str, port: int) -> str: ...
    def sendeprt(self, host: str, port: int) -> str: ...
    def makeport(self) -> socket: ...
    def makepasv(self) -> tuple[str, int]: ...
    def login(self, user: str = ..., passwd: str = ..., acct: str = ...) -> str: ...
    # In practice, `rest` rest can actually be anything whose str() is an integer sequence, so to make it simple we allow integers.
    def ntransfercmd(self, cmd: str, rest: int | str | None = ...) -> tuple[socket, int]: ...
    def transfercmd(self, cmd: str, rest: int | str | None = ...) -> socket: ...
    def retrbinary(
        self, cmd: str, callback: Callable[[bytes], Any], blocksize: int = ..., rest: int | str | None = ...
    ) -> str: ...
    def storbinary(
        self,
        cmd: str,
        fp: SupportsRead[bytes],
        blocksize: int = ...,
        callback: Callable[[bytes], Any] | None = ...,
        rest: int | str | None = ...,
    ) -> str: ...
    def retrlines(self, cmd: str, callback: Callable[[str], Any] | None = ...) -> str: ...
    def storlines(self, cmd: str, fp: SupportsReadline[bytes], callback: Callable[[bytes], Any] | None = ...) -> str: ...
    def acct(self, password: str) -> str: ...
    def nlst(self, *args: str) -> list[str]: ...
    # Technically only the last arg can be a Callable but ...
    def dir(self, *args: str | Callable[[str], None]) -> None: ...
    def mlsd(self, path: str = ..., facts: Iterable[str] = ...) -> Iterator[tuple[str, dict[str, str]]]: ...
    def rename(self, fromname: str, toname: str) -> str: ...
    def delete(self, filename: str) -> str: ...
    def cwd(self, dirname: str) -> str: ...
    def size(self, filename: str) -> int | None: ...
    def mkd(self, dirname: str) -> str: ...
    def rmd(self, dirname: str) -> str: ...
    def pwd(self) -> str: ...
    def quit(self) -> str: ...
    def close(self) -> None: ...

class FTP_TLS(FTP):
    if sys.version_info >= (3, 9):
        def __init__(
            self,
            host: str = ...,
            user: str = ...,
            passwd: str = ...,
            acct: str = ...,
            keyfile: str | None = ...,
            certfile: str | None = ...,
            context: SSLContext | None = ...,
            timeout: float = ...,
            source_address: tuple[str, int] | None = ...,
            *,
            encoding: str = ...,
        ) -> None: ...
    else:
        def __init__(
            self,
            host: str = ...,
            user: str = ...,
            passwd: str = ...,
            acct: str = ...,
            keyfile: str | None = ...,
            certfile: str | None = ...,
            context: SSLContext | None = ...,
            timeout: float = ...,
            source_address: tuple[str, int] | None = ...,
        ) -> None: ...
    ssl_version: int
    keyfile: str | None
    certfile: str | None
    context: SSLContext
    def login(self, user: str = ..., passwd: str = ..., acct: str = ..., secure: bool = ...) -> str: ...
    def auth(self) -> str: ...
    def prot_p(self) -> str: ...
    def prot_c(self) -> str: ...
    def ccc(self) -> str: ...

def parse150(resp: str) -> int | None: ...  # undocumented
def parse227(resp: str) -> tuple[str, int]: ...  # undocumented
def parse229(resp: str, peer: Any) -> tuple[str, int]: ...  # undocumented
def parse257(resp: str) -> str: ...  # undocumented
def ftpcp(
    source: FTP, sourcename: str, target: FTP, targetname: str = ..., type: Literal["A", "I"] = ...
) -> None: ...  # undocumented



from _curses import _CursesWindow
from typing import Callable

def rectangle(win: _CursesWindow, uly: int, ulx: int, lry: int, lrx: int) -> None: ...

class Textbox:
    stripspaces: bool
    def __init__(self, win: _CursesWindow, insert_mode: bool = ...) -> None: ...
    def edit(self, validate: Callable[[int], int] | None = ...) -> str: ...
    def do_command(self, ch: str | int) -> None: ...
    def gather(self) -> str: ...



from typing import Any

KEY: str
POINTER: str
NONE: str
SOURCE_TYPES: Any
POINTER_MOUSE: str
POINTER_TOUCH: str
POINTER_PEN: str
POINTER_KINDS: Any

class Interaction:
    PAUSE: str
    source: Any
    def __init__(self, source) -> None: ...

class Pause(Interaction):
    source: Any
    duration: Any
    def __init__(self, source, duration: int = ...) -> None: ...
    def encode(self): ...



from _typeshed import Self
from types import TracebackType
from typing import Iterator, MutableMapping
from typing_extensions import Literal

__all__ = ["open", "whichdb", "error"]

_KeyType = str | bytes
_ValueType = str | bytes
_TFlags = Literal[
    "r",
    "w",
    "c",
    "n",
    "rf",
    "wf",
    "cf",
    "nf",
    "rs",
    "ws",
    "cs",
    "ns",
    "ru",
    "wu",
    "cu",
    "nu",
    "rfs",
    "wfs",
    "cfs",
    "nfs",
    "rfu",
    "wfu",
    "cfu",
    "nfu",
    "rsf",
    "wsf",
    "csf",
    "nsf",
    "rsu",
    "wsu",
    "csu",
    "nsu",
    "ruf",
    "wuf",
    "cuf",
    "nuf",
    "rus",
    "wus",
    "cus",
    "nus",
    "rfsu",
    "wfsu",
    "cfsu",
    "nfsu",
    "rfus",
    "wfus",
    "cfus",
    "nfus",
    "rsfu",
    "wsfu",
    "csfu",
    "nsfu",
    "rsuf",
    "wsuf",
    "csuf",
    "nsuf",
    "rufs",
    "wufs",
    "cufs",
    "nufs",
    "rusf",
    "wusf",
    "cusf",
    "nusf",
]

class _Database(MutableMapping[_KeyType, bytes]):
    def close(self) -> None: ...
    def __getitem__(self, key: _KeyType) -> bytes: ...
    def __setitem__(self, key: _KeyType, value: _ValueType) -> None: ...
    def __delitem__(self, key: _KeyType) -> None: ...
    def __iter__(self) -> Iterator[bytes]: ...
    def __len__(self) -> int: ...
    def __del__(self) -> None: ...
    def __enter__(self: Self) -> Self: ...
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None: ...

class _error(Exception): ...

error: tuple[type[_error], type[OSError]]

def whichdb(filename: str) -> str: ...
def open(file: str, flag: _TFlags = ..., mode: int = ...) -> _Database: ...



from typing import Any, Callable, Hashable, SupportsInt, TypeVar, Union

_T = TypeVar("_T")
_Reduce = Union[tuple[Callable[..., _T], tuple[Any, ...]], tuple[Callable[..., _T], tuple[Any, ...], Any | None]]

__all__ = ["pickle", "constructor", "add_extension", "remove_extension", "clear_extension_cache"]

def pickle(
    ob_type: type[_T],
    pickle_function: Callable[[_T], str | _Reduce[_T]],
    constructor_ob: Callable[[_Reduce[_T]], _T] | None = ...,
) -> None: ...
def constructor(object: Callable[[_Reduce[_T]], _T]) -> None: ...
def add_extension(module: Hashable, name: Hashable, code: SupportsInt) -> None: ...
def remove_extension(module: Hashable, name: Hashable, code: int) -> None: ...
def clear_extension_cache() -> None: ...

_DispatchTableType = dict[type, Callable[[Any], str | _Reduce[Any]]]  # imported by multiprocessing.reduction
dispatch_table: _DispatchTableType  # undocumented



from distutils.core import Command
from typing import Any

class dist_info(Command):
    description: str
    user_options: Any
    egg_base: Any
    def initialize_options(self) -> None: ...
    def finalize_options(self) -> None: ...
    def run(self) -> None: ...



from distutils.cmd import Command
from typing import ClassVar

class install_egg_info(Command):
    description: ClassVar[str]
    user_options: ClassVar[list[tuple[str, str | None, str]]]
    def initialize_options(self) -> None: ...
    def finalize_options(self) -> None: ...
    def run(self) -> None: ...
    def get_outputs(self) -> list[str]: ...



from typing import Any

class CommandsParser:
    initialized: bool
    commands: Any
    def __init__(self, redis_connection) -> None: ...
    def initialize(self, r) -> None: ...
    def get_keys(self, redis_conn, *args): ...



"""Generate cross references from a project."""

import signal
import sys
import traceback

from pytype import utils

from pytype.tools.xref import debug
from pytype.tools.xref import indexer
from pytype.tools.xref import kythe
from pytype.tools.xref import output
from pytype.tools.xref import parse_args


def main():
  try:
    args, kythe_args, options = parse_args.parse_args(sys.argv[1:])
  except utils.UsageError as e:
    print(str(e), file=sys.stderr)
    sys.exit(1)

  if options.timeout is not None:
    signal.alarm(options.timeout)

  try:
    ix = indexer.process_file(options, generate_callgraphs=True)
  except indexer.PytypeError as e:
    print(e.args[0], file=sys.stderr)
    if args.debug:
      traceback.print_exc()
    else:
      print("Run with --debug to see a traceback.")
    sys.exit(1)

  if args.debug:
    debug.show_index(ix)
  else:
    kythe_graph = kythe.generate_graph(ix, kythe_args)
    output.output_kythe_graph(kythe_graph)


if __name__ == "__main__":
  sys.exit(main())



from typing import Any

from . import base

class Filter(base.Filter):
    require_matching_tags: Any
    def __init__(self, source, require_matching_tags: bool = ...) -> None: ...
    def __iter__(self): ...



"""Function definitions in pyi files."""

import sys
import textwrap

from typing import Any, List, Optional

from pytype import utils
from pytype.pyi import types
from pytype.pyi.types import ParseError  # pylint: disable=g-importing-member
from pytype.pytd import pytd
from pytype.pytd import visitors
from pytype.pytd.codegen import function as pytd_function

# pylint: disable=g-import-not-at-top
if sys.version_info >= (3, 8):
  import ast as ast3
else:
  from typed_ast import ast3
# pylint: enable=g-import-not-at-top


class Mutator(visitors.Visitor):
  """Visitor for adding a mutated_type to parameters.

  We model the parameter x in
    def f(x: old_type):
      x = new_type
  as
    Parameter(name=x, type=old_type, mutated_type=new_type)
  .
  This visitor applies the body "x = new_type" to the function signature.
  """

  def __init__(self, name, new_type):
    super().__init__()
    self.name = name
    self.new_type = new_type
    self.successful = False

  def VisitParameter(self, p):
    if p.name == self.name:
      self.successful = True
      if p.optional:
        raise NotImplementedError(
            f"Argument {p.name!r} cannot be both mutable and optional")
      return p.Replace(mutated_type=self.new_type)
    else:
      return p

  def __repr__(self):
    return f"Mutator<{self.name} -> {self.new_type}>"

  __str__ = __repr__


class Param(pytd_function.Param):
  """Internal representation of function parameters."""

  @classmethod
  def from_arg(cls, arg: ast3.arg, kind: pytd.ParameterKind) -> "Param":
    """Constructor from an ast.argument node."""
    p = cls(arg.arg)
    if arg.annotation:
      p.type = arg.annotation
    p.kind = kind
    return p


class NameAndSig(pytd_function.NameAndSig):
  """Internal representation of function signatures."""

  @classmethod
  def from_function(
      cls, function: ast3.FunctionDef, is_async: bool) -> "NameAndSig":
    """Constructor from an ast.FunctionDef node."""
    name = function.name

    # decorators
    decorators = set(function.decorator_list)
    abstracts = {"abstractmethod", "abc.abstractmethod"}
    coroutines = {"coroutine", "asyncio.coroutine", "coroutines.coroutine"}
    overload = {"overload"}
    final = {"final"}
    ignored = {"type_check_only"}
    is_abstract = bool(decorators & abstracts)
    is_coroutine = bool(decorators & coroutines)
    is_overload = bool(decorators & overload)
    is_final = bool(decorators & final)
    decorators -= (abstracts | coroutines | overload | final | ignored)
    # TODO(mdemello): do we need this limitation?
    if len(decorators) > 1:
      raise ParseError(f"Too many decorators for {name}")
    decorator, = decorators if decorators else (None,)

    exceptions = []
    mutators = []
    for i, x in enumerate(function.body):
      if isinstance(x, types.Raise):
        exceptions.append(x.exception)
      elif isinstance(x, Mutator):
        mutators.append(x)
      elif isinstance(x, types.Ellipsis):
        pass
      elif (isinstance(x, ast3.Expr) and
            isinstance(x.value, ast3.Str) and
            i == 0):
        # docstring
        pass
      else:
        msg = textwrap.dedent("""
            Unexpected statement in function body.
            Only `raise` statements and type mutations are valid
        """).lstrip()
        if isinstance(x, ast3.AST):
          raise ParseError(msg).at(x)
        else:
          raise ParseError(msg)

    # exceptions
    sig = _pytd_signature(function, is_async, exceptions=exceptions)

    # mutators
    # If `self` is generic, a type parameter is being mutated.
    if (sig.params and sig.params[0].name == "self" and
        isinstance(sig.params[0].type, pytd.GenericType)):
      mutators.append(Mutator("self", sig.params[0].type))
    for mutator in mutators:
      try:
        sig = sig.Visit(mutator)
      except NotImplementedError as e:
        raise ParseError(utils.message(e)) from e
      if not mutator.successful:
        raise ParseError(f"No parameter named {mutator.name!r}")

    return cls(name, sig, decorator, is_abstract, is_coroutine, is_final,
               is_overload)


def _pytd_signature(
    function: ast3.FunctionDef,
    is_async: bool,
    exceptions: Optional[List[pytd.Type]] = None
) -> pytd.Signature:
  """Construct a pytd signature from an ast.FunctionDef node."""
  name = function.name
  args = function.args
  # Positional-only parameters are new in Python 3.8.
  posonly_params = [Param.from_arg(a, pytd.ParameterKind.POSONLY)
                    for a in getattr(args, "posonlyargs", ())]
  pos_params = [Param.from_arg(a, pytd.ParameterKind.REGULAR)
                for a in args.args]
  kwonly_params = [Param.from_arg(a, pytd.ParameterKind.KWONLY)
                   for a in args.kwonlyargs]
  _apply_defaults(posonly_params + pos_params, args.defaults)
  _apply_defaults(kwonly_params, args.kw_defaults)
  all_params = posonly_params + pos_params + kwonly_params
  params = tuple(x.to_pytd() for x in all_params)
  starargs = _pytd_star_param(args.vararg)
  starstarargs = _pytd_starstar_param(args.kwarg)
  ret = pytd_function.pytd_return_type(name, function.returns, is_async)
  exceptions = exceptions or []
  return pytd.Signature(params=params,
                        return_type=ret,
                        starargs=starargs,
                        starstarargs=starstarargs,
                        exceptions=tuple(exceptions), template=())


def _pytd_star_param(arg: ast3.arg) -> Optional[pytd.Parameter]:
  """Return a pytd.Parameter for a *args argument."""
  if not arg:
    return None
  return pytd_function.pytd_star_param(arg.arg, arg.annotation)


def _pytd_starstar_param(
    arg: Optional[ast3.arg]) -> Optional[pytd.Parameter]:
  """Return a pytd.Parameter for a **kwargs argument."""
  if not arg:
    return None
  return pytd_function.pytd_starstar_param(arg.arg, arg.annotation)


def _apply_defaults(params: List[Param], defaults: List[Any]) -> None:
  for p, d in zip(reversed(params), reversed(defaults)):
    if d is None:
      continue
    elif isinstance(d, types.Pyval):
      p.default = d.to_pytd()
    else:
      p.default = pytd.AnythingType()



"""Utilities for working with pytype source files."""

# All functions with a dependence on __file__ should go here.
# This file should be kept under pytype/ so that __file__.dirname is the
# top-level pytype directory.

import os
import re


class NoSuchDirectory(Exception):  # pylint: disable=g-bad-exception-name
  pass


def pytype_source_dir():
  """The base directory of the pytype source tree."""
  res = os.path.dirname(__file__)
  if os.path.basename(res) == "__pycache__":
    # For source-less par files __file__ points at __pycache__ subdirectory...
    res = os.path.dirname(res)
  return res


def get_full_path(path):
  """Full path to a file or directory within the pytype source tree.

  Arguments:
    path: An absolute or relative path.

  Returns:
    path for absolute paths.
    full path resolved relative to pytype/ for relative paths.
  """
  # TODO(mdemello): Insist on relative paths here.
  return os.path.join(pytype_source_dir(), path)


def load_text_file(filename):
  return _load_data_file(filename, text=True)


def load_binary_file(filename):
  return _load_data_file(filename, text=False)


def _load_data_file(filename, text):
  """Get the contents of a data file from the pytype installation.

  Arguments:
    filename: the path, relative to "pytype/"
    text: whether to load the file as text or bytes.
  Returns:
    The contents of the file as a bytestring
  Raises:
    IOError: if file not found
  """
  path = filename if os.path.isabs(filename) else  get_full_path(filename)
  # Check for a ResourceLoader (see comment under list_pytype_files).
  loader = globals().get("__loader__", None)
  if loader:
    # For an explanation of the args to loader.get_data, see
    # https://www.python.org/dev/peps/pep-0302/#optional-extensions-to-the-importer-protocol
    # https://docs.python.org/3/library/importlib.html#importlib.abc.ResourceLoader.get_data
    data = loader.get_data(path)
    if text:
      # Opening a file in text mode automatically converts carriage returns to
      # newlines, so we manually simulate that behavior here.
      return re.sub("\r\n?", "\n", data.decode("utf-8"))
    return data
  with open(path, "r" if text else "rb") as fi:
    return fi.read()


def list_files(basedir):
  """List files in the directory rooted at |basedir|."""
  if not os.path.isdir(basedir):
    raise NoSuchDirectory(basedir)
  directories = [""]
  while directories:
    d = directories.pop()
    for basename in os.listdir(os.path.join(basedir, d)):
      filename = os.path.join(d, basename)
      if os.path.isdir(os.path.join(basedir, filename)):
        directories.append(filename)
      elif os.path.exists(os.path.join(basedir, filename)):
        yield filename


def list_pytype_files(suffix):
  """Recursively get the contents of a directory in the pytype installation.

  This reports files in said directory as well as all subdirectories of it.

  Arguments:
    suffix: the path, relative to "pytype/"
  Yields:
    The filenames, relative to pytype/{suffix}
  Raises:
    NoSuchDirectory: if the directory doesn't exist.
  """
  assert not suffix.endswith("/")
  loader = globals().get("__loader__", None)
  try:
    # List directory using __loader__.
    # __loader__ exists only when this file is in a Python archive in Python 2
    # but is always present in Python 3, so we can't use the presence or
    # absence of the loader to determine whether calling get_zipfile is okay.
    filenames = loader.get_zipfile().namelist()  # pytype: disable=attribute-error
  except AttributeError:
    # List directory using the file system
    yield from list_files(get_full_path(suffix))
  else:
    for filename in filenames:
      directory = "pytype/" + suffix + "/"
      try:
        i = filename.rindex(directory)
      except ValueError:
        pass
      else:
        yield filename[i + len(directory):]



import importlib.abc
import sys
import types
from typing import Any, Callable, Iterable, Sequence

if sys.version_info >= (3, 8):
    from importlib.metadata import DistributionFinder, PathDistribution

class ModuleSpec:
    def __init__(
        self,
        name: str,
        loader: importlib.abc.Loader | None,
        *,
        origin: str | None = ...,
        loader_state: Any = ...,
        is_package: bool | None = ...,
    ) -> None: ...
    name: str
    loader: importlib.abc.Loader | None
    origin: str | None
    submodule_search_locations: list[str] | None
    loader_state: Any
    cached: str | None
    @property
    def parent(self) -> str | None: ...
    has_location: bool
    def __eq__(self, other: object) -> bool: ...

class BuiltinImporter(importlib.abc.MetaPathFinder, importlib.abc.InspectLoader):
    # MetaPathFinder
    @classmethod
    def find_module(cls, fullname: str, path: Sequence[importlib.abc._Path] | None = ...) -> importlib.abc.Loader | None: ...
    @classmethod
    def find_spec(
        cls, fullname: str, path: Sequence[importlib.abc._Path] | None = ..., target: types.ModuleType | None = ...
    ) -> ModuleSpec | None: ...
    # InspectLoader
    @classmethod
    def is_package(cls, fullname: str) -> bool: ...
    @classmethod
    def load_module(cls, fullname: str) -> types.ModuleType: ...
    @classmethod
    def get_code(cls, fullname: str) -> None: ...
    @classmethod
    def get_source(cls, fullname: str) -> None: ...
    # Loader
    @staticmethod
    def module_repr(module: types.ModuleType) -> str: ...
    if sys.version_info >= (3, 10):
        @staticmethod
        def create_module(spec: ModuleSpec) -> types.ModuleType | None: ...
        @staticmethod
        def exec_module(module: types.ModuleType) -> None: ...
    else:
        @classmethod
        def create_module(cls, spec: ModuleSpec) -> types.ModuleType | None: ...
        @classmethod
        def exec_module(cls, module: types.ModuleType) -> None: ...

class FrozenImporter(importlib.abc.MetaPathFinder, importlib.abc.InspectLoader):
    # MetaPathFinder
    @classmethod
    def find_module(cls, fullname: str, path: Sequence[importlib.abc._Path] | None = ...) -> importlib.abc.Loader | None: ...
    @classmethod
    def find_spec(
        cls, fullname: str, path: Sequence[importlib.abc._Path] | None = ..., target: types.ModuleType | None = ...
    ) -> ModuleSpec | None: ...
    # InspectLoader
    @classmethod
    def is_package(cls, fullname: str) -> bool: ...
    @classmethod
    def load_module(cls, fullname: str) -> types.ModuleType: ...
    @classmethod
    def get_code(cls, fullname: str) -> None: ...
    @classmethod
    def get_source(cls, fullname: str) -> None: ...
    # Loader
    @staticmethod
    def module_repr(m: types.ModuleType) -> str: ...
    if sys.version_info >= (3, 10):
        @staticmethod
        def create_module(spec: ModuleSpec) -> types.ModuleType | None: ...
    else:
        @classmethod
        def create_module(cls, spec: ModuleSpec) -> types.ModuleType | None: ...

    @staticmethod
    def exec_module(module: types.ModuleType) -> None: ...

class WindowsRegistryFinder(importlib.abc.MetaPathFinder):
    @classmethod
    def find_module(cls, fullname: str, path: Sequence[importlib.abc._Path] | None = ...) -> importlib.abc.Loader | None: ...
    @classmethod
    def find_spec(
        cls, fullname: str, path: Sequence[importlib.abc._Path] | None = ..., target: types.ModuleType | None = ...
    ) -> ModuleSpec | None: ...

class PathFinder:
    if sys.version_info >= (3, 10):
        @staticmethod
        def invalidate_caches() -> None: ...
    else:
        @classmethod
        def invalidate_caches(cls) -> None: ...
    if sys.version_info >= (3, 10):
        @staticmethod
        def find_distributions(context: DistributionFinder.Context = ...) -> Iterable[PathDistribution]: ...
    elif sys.version_info >= (3, 8):
        @classmethod
        def find_distributions(cls, context: DistributionFinder.Context = ...) -> Iterable[PathDistribution]: ...

    @classmethod
    def find_spec(
        cls, fullname: str, path: Sequence[bytes | str] | None = ..., target: types.ModuleType | None = ...
    ) -> ModuleSpec | None: ...
    @classmethod
    def find_module(cls, fullname: str, path: Sequence[bytes | str] | None = ...) -> importlib.abc.Loader | None: ...

SOURCE_SUFFIXES: list[str]
DEBUG_BYTECODE_SUFFIXES: list[str]
OPTIMIZED_BYTECODE_SUFFIXES: list[str]
BYTECODE_SUFFIXES: list[str]
EXTENSION_SUFFIXES: list[str]

def all_suffixes() -> list[str]: ...

class FileFinder(importlib.abc.PathEntryFinder):
    path: str
    def __init__(self, path: str, *loader_details: tuple[type[importlib.abc.Loader], list[str]]) -> None: ...
    @classmethod
    def path_hook(
        cls, *loader_details: tuple[type[importlib.abc.Loader], list[str]]
    ) -> Callable[[str], importlib.abc.PathEntryFinder]: ...

class SourceFileLoader(importlib.abc.FileLoader, importlib.abc.SourceLoader):
    def set_data(self, path: importlib.abc._Path, data: bytes, *, _mode: int = ...) -> None: ...

class SourcelessFileLoader(importlib.abc.FileLoader, importlib.abc.SourceLoader): ...

class ExtensionFileLoader(importlib.abc.ExecutionLoader):
    def __init__(self, name: str, path: importlib.abc._Path) -> None: ...
    def get_filename(self, name: str | None = ...) -> importlib.abc._Path: ...
    def get_source(self, fullname: str) -> None: ...
    def create_module(self, spec: ModuleSpec) -> types.ModuleType: ...
    def exec_module(self, module: types.ModuleType) -> None: ...
    def is_package(self, fullname: str) -> bool: ...
    def get_code(self, fullname: str) -> None: ...
    def __eq__(self, other: object) -> bool: ...



from typing import Any, Iterable, Iterator

from .key import Key

def bucket_lister(
    bucket, prefix: str = ..., delimiter: str = ..., marker: str = ..., headers: Any | None = ..., encoding_type: Any | None = ...
): ...

class BucketListResultSet(Iterable[Key]):
    bucket: Any
    prefix: Any
    delimiter: Any
    marker: Any
    headers: Any
    encoding_type: Any
    def __init__(
        self,
        bucket: Any | None = ...,
        prefix: str = ...,
        delimiter: str = ...,
        marker: str = ...,
        headers: Any | None = ...,
        encoding_type: Any | None = ...,
    ) -> None: ...
    def __iter__(self) -> Iterator[Key]: ...

def versioned_bucket_lister(
    bucket,
    prefix: str = ...,
    delimiter: str = ...,
    key_marker: str = ...,
    version_id_marker: str = ...,
    headers: Any | None = ...,
    encoding_type: Any | None = ...,
): ...

class VersionedBucketListResultSet:
    bucket: Any
    prefix: Any
    delimiter: Any
    key_marker: Any
    version_id_marker: Any
    headers: Any
    encoding_type: Any
    def __init__(
        self,
        bucket: Any | None = ...,
        prefix: str = ...,
        delimiter: str = ...,
        key_marker: str = ...,
        version_id_marker: str = ...,
        headers: Any | None = ...,
        encoding_type: Any | None = ...,
    ) -> None: ...
    def __iter__(self) -> Iterator[Key]: ...

def multipart_upload_lister(
    bucket, key_marker: str = ..., upload_id_marker: str = ..., headers: Any | None = ..., encoding_type: Any | None = ...
): ...

class MultiPartUploadListResultSet:
    bucket: Any
    key_marker: Any
    upload_id_marker: Any
    headers: Any
    encoding_type: Any
    def __init__(
        self,
        bucket: Any | None = ...,
        key_marker: str = ...,
        upload_id_marker: str = ...,
        headers: Any | None = ...,
        encoding_type: Any | None = ...,
    ) -> None: ...
    def __iter__(self): ...



"""Utilities for parsing pytd files for builtins."""

import os

from pytype.pyi import parser
from pytype.pytd import pytd_utils
from pytype.pytd import visitors


# TODO(rechen): It would be nice to get rid of GetBuiltinsAndTyping, and let the
# loader call BuiltinsAndTyping.load directly, but the cache currently prevents
# slowdowns in tests that create loaders willy-nilly.  Maybe load_pytd.py can
# warn if there are more than n loaders in play, at any given time.
_cached_builtins_pytd = []

# pylint: disable=invalid-name
# We use a mix of camel case and snake case method names in this file.


def InvalidateCache():
  if _cached_builtins_pytd:
    del _cached_builtins_pytd[0]


# Do not call this - get the "builtins" and "typing" modules via the loader.
def GetBuiltinsAndTyping(options):
  if not _cached_builtins_pytd:
    _cached_builtins_pytd.append(BuiltinsAndTyping().load(options))
  return _cached_builtins_pytd[0]


# pyi for a catch-all module
DEFAULT_SRC = """
from typing import Any
def __getattr__(name: Any) -> Any: ...
"""


# If you have a Loader available, use loader.get_default_ast() instead.
def GetDefaultAst(options):
  return parser.parse_string(src=DEFAULT_SRC, options=options)


class BuiltinsAndTyping:
  """The builtins and typing modules, which need to be treated specially."""

  def _parse_predefined(self, name, options):
    _, src = pytd_utils.GetPredefinedFile("builtins", name, ".pytd")
    mod = parser.parse_string(src, name=name, options=options)
    return mod

  def load(self, options):
    """Read builtins.pytd and typing.pytd, and return the parsed modules."""
    t = self._parse_predefined("typing", options)
    b = self._parse_predefined("builtins", options)
    b = b.Visit(visitors.LookupExternalTypes({"typing": t},
                                             self_name="builtins"))
    t = t.Visit(visitors.LookupBuiltins(b))
    b = b.Visit(visitors.NamedTypeToClassType())
    t = t.Visit(visitors.NamedTypeToClassType())
    b = b.Visit(visitors.AdjustTypeParameters())
    t = t.Visit(visitors.AdjustTypeParameters())
    b = b.Visit(visitors.CanonicalOrderingVisitor())
    t = t.Visit(visitors.CanonicalOrderingVisitor())
    b.Visit(visitors.FillInLocalPointers({"": b, "typing": t,
                                          "builtins": b}))
    t.Visit(visitors.FillInLocalPointers({"": t, "typing": t,
                                          "builtins": b}))
    b.Visit(visitors.VerifyLookup())
    t.Visit(visitors.VerifyLookup())
    b.Visit(visitors.VerifyContainers())
    t.Visit(visitors.VerifyContainers())
    return b, t


class BuiltinLoader:
  """Load builtins from the pytype source tree."""

  def __init__(self, options):
    self.options = options

  def _parse_predefined(self, pytd_subdir, module, as_package=False):
    """Parse a pyi/pytd file in the pytype source tree."""
    try:
      filename, src = pytd_utils.GetPredefinedFile(
          pytd_subdir, module, as_package=as_package)
    except OSError:
      return None
    ast = parser.parse_string(
        src, filename=filename, name=module, options=self.options)
    assert ast.name == module
    return ast

  def get_builtin(self, builtin_dir, module_name):
    """Load a stub that ships with pytype."""
    mod = self._parse_predefined(builtin_dir, module_name)
    # For stubs in pytype's stubs/ directory, we use the module name prefixed
    # with "pytd:" for the filename. Package filenames need an "/__init__.pyi"
    # suffix for Module.is_package to recognize them.
    if mod:
      filename = module_name
    else:
      mod = self._parse_predefined(builtin_dir, module_name, as_package=True)
      filename = os.path.join(module_name, "__init__.pyi")
    return filename, mod

# pylint: enable=invalid-name



"""Tests for arg_parser."""

import argparse
import types

from pytype import config as pytype_config
from pytype import datatypes
from pytype.tools import arg_parser

import unittest


class TestConvertString(unittest.TestCase):
  """Test arg_parser.convert_string."""

  def test_int(self):
    self.assertEqual(arg_parser.convert_string('3'), 3)

  def test_bool(self):
    self.assertIs(arg_parser.convert_string('True'), True)
    self.assertIs(arg_parser.convert_string('False'), False)

  def test_whitespace(self):
    self.assertEqual(arg_parser.convert_string('err1,\nerr2'), 'err1,err2')


def make_parser():
  """Construct a parser to run tests against."""

  parser = argparse.ArgumentParser()
  parser.add_argument(
      '-v', '--verbosity', dest='verbosity', type=int, action='store',
      default=1)
  parser.add_argument(
      '--config', dest='config', type=str, action='store', default='')

  # Add options from pytype-single.
  wrapper = datatypes.ParserWrapper(parser)
  pytype_config.add_basic_options(wrapper)
  return arg_parser.Parser(parser, wrapper.actions)


class TestParser(unittest.TestCase):
  """Test arg_parser.Parser."""

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.parser = make_parser()

  def test_verbosity(self):
    self.assertEqual(self.parser.parse_args(['--verbosity', '0']).verbosity, 0)
    self.assertEqual(self.parser.parse_args(['-v1']).verbosity, 1)

  def test_config(self):
    args = self.parser.parse_args(['--config=test.cfg'])
    self.assertEqual(args.config, 'test.cfg')

  def test_pytype_single_args(self):
    args = self.parser.parse_args(['--disable=import-error'])
    self.assertSequenceEqual(args.disable, ['import-error'])

  def test_postprocess(self):
    args = types.SimpleNamespace(disable='import-error')
    self.parser.postprocess(args)
    self.assertSequenceEqual(args.disable, ['import-error'])

  def test_postprocess_from_strings(self):
    args = types.SimpleNamespace(report_errors='False', protocols='True')
    self.parser.postprocess(args, from_strings=True)
    self.assertFalse(args.report_errors)
    self.assertTrue(args.protocols)


if __name__ == '__main__':
  unittest.main()



"""Tests for module_utils.py."""

import os

from pytype import file_utils
from pytype import module_utils

import unittest


class ModuleUtilsTest(unittest.TestCase):
  """Test module utilities."""

  def test_get_absolute_name(self):
    test_cases = [
        ("x.y", "a.b", "x.y.a.b"),
        ("", "a.b", "a.b"),
        ("x.y", ".a.b", "x.y.a.b"),
        ("x.y", "..a.b", "x.a.b"),
        ("x.y", "...a.b", None),
    ]
    for prefix, name, expected in test_cases:
      self.assertEqual(module_utils.get_absolute_name(prefix, name), expected)

  def test_path_to_module_name(self):
    self.assertIsNone(module_utils.path_to_module_name("../foo.py"))
    self.assertIsNone(module_utils.path_to_module_name("x/y/foo.txt"))
    self.assertEqual("x.y.z", module_utils.path_to_module_name("x/y/z.pyi"))
    self.assertEqual("x.y.z", module_utils.path_to_module_name("x/y/z.pytd"))
    self.assertEqual("x.y.z",
                     module_utils.path_to_module_name("x/y/z/__init__.pyi"))


# Because TestInferModule expands a lot of paths:
expand = file_utils.expand_path


class TestInferModule(unittest.TestCase):
  """Test module_utils.infer_module."""

  def assert_module_equal(self, module, path, target, name, kind="Local"):
    self.assertEqual(module.path.rstrip(os.sep), path.rstrip(os.sep))
    self.assertEqual(module.target, target)
    self.assertEqual(module.name, name)
    self.assertEqual(module.kind, kind)

  def test_simple_name(self):
    mod = module_utils.infer_module(expand("foo/bar.py"), [expand("foo")])
    self.assert_module_equal(mod, expand("foo"), "bar.py", "bar")

  def test_name_in_package(self):
    mod = module_utils.infer_module(expand("foo/bar/baz.py"), [expand("foo")])
    self.assert_module_equal(mod, expand("foo"), "bar/baz.py", "bar.baz")

  def test_multiple_paths(self):
    pythonpath = [expand("foo"), expand("bar/baz"), expand("bar")]
    mod = module_utils.infer_module(expand("bar/baz/qux.py"), pythonpath)
    self.assert_module_equal(mod, expand("bar/baz"), "qux.py", "qux")
    mod = module_utils.infer_module(expand("bar/qux.py"), pythonpath)
    self.assert_module_equal(mod, expand("bar"), "qux.py", "qux")

  def test_not_found(self):
    mod = module_utils.infer_module(expand("bar/baz.py"), ["foo"])
    expected_target = expand("bar/baz.py")
    expected_name, _ = os.path.splitext(expected_target.replace(os.sep, "."))
    self.assert_module_equal(mod, "", expected_target, expected_name)


if __name__ == "__main__":
  unittest.main()



from typing import Callable, Iterable, Mapping

def is_array(checker, instance) -> bool: ...
def is_bool(checker, instance) -> bool: ...
def is_integer(checker, instance) -> bool: ...
def is_null(checker, instance) -> bool: ...
def is_number(checker, instance) -> bool: ...
def is_object(checker, instance) -> bool: ...
def is_string(checker, instance) -> bool: ...
def is_any(checker, instance) -> bool: ...

class TypeChecker:
    def __init__(self, type_checkers: Mapping[str, Callable[[object], bool]] = ...) -> None: ...
    def is_type(self, instance, type: str) -> bool: ...
    def redefine(self, type: str, fn: Callable[..., bool]) -> TypeChecker: ...
    def redefine_many(self, definitions=...) -> TypeChecker: ...
    def remove(self, *types: Iterable[str]) -> TypeChecker: ...

draft3_type_checker: TypeChecker
draft4_type_checker: TypeChecker
draft6_type_checker: TypeChecker
draft7_type_checker: TypeChecker
draft201909_type_checker: TypeChecker
draft202012_type_checker: TypeChecker



class Reporter:
    def __init__(self, warningStream, errorStream) -> None: ...
    def unexpectedError(self, filename, msg) -> None: ...
    def syntaxError(self, filename, msg, lineno, offset, text) -> None: ...
    def flake(self, message) -> None: ...



from typing import Any

from ...sql import expression
from ...types import TypeDecorator

ischema: Any

class CoerceUnicode(TypeDecorator):
    impl: Any
    cache_ok: bool
    def process_bind_param(self, value, dialect): ...
    def bind_expression(self, bindvalue): ...

class _cast_on_2005(expression.ColumnElement[Any]):
    bindvalue: Any
    def __init__(self, bindvalue) -> None: ...

schemata: Any
tables: Any
columns: Any
mssql_temp_table_columns: Any
constraints: Any
column_constraints: Any
key_constraints: Any
ref_constraints: Any
views: Any
computed_columns: Any
sequences: Any

class IdentitySqlVariant(TypeDecorator):
    impl: Any
    cache_ok: bool
    def column_expression(self, colexpr): ...

identity_columns: Any



# requests also imports urllib3 as requests.packages.urllib3, the stubs don't reflect that.

class VendorAlias:
    def __init__(self, package_names) -> None: ...
    def find_module(self, fullname, path=...): ...
    def load_module(self, name): ...



from _typeshed import Self
from typing import Any, Mapping

from .retry import Retry

ssl_available: Any
SYM_STAR: Any
SYM_DOLLAR: Any
SYM_CRLF: Any
SYM_EMPTY: Any
SERVER_CLOSED_CONNECTION_ERROR: Any

# Options as passed to Pool.get_connection().
_ConnectionPoolOptions = Any

class BaseParser:
    EXCEPTION_CLASSES: Any
    def parse_error(self, response): ...

class SocketBuffer:
    socket_read_size: Any
    bytes_written: Any
    bytes_read: Any
    def __init__(self, socket, socket_read_size, socket_timeout) -> None: ...
    @property
    def length(self): ...
    def read(self, length): ...
    def readline(self): ...
    def purge(self): ...
    def close(self): ...
    def can_read(self, timeout): ...

class PythonParser(BaseParser):
    encoding: Any
    socket_read_size: Any
    def __init__(self, socket_read_size) -> None: ...
    def __del__(self): ...
    def on_connect(self, connection): ...
    def on_disconnect(self): ...
    def can_read(self, timeout): ...
    def read_response(self, disable_decoding: bool = ...): ...

class HiredisParser(BaseParser):
    socket_read_size: Any
    def __init__(self, socket_read_size) -> None: ...
    def __del__(self): ...
    def on_connect(self, connection, **kwargs): ...
    def on_disconnect(self): ...
    def can_read(self, timeout): ...
    def read_from_socket(self, timeout=..., raise_on_timeout: bool = ...) -> bool: ...
    def read_response(self, disable_decoding: bool = ...): ...

DefaultParser: Any

class Encoder:
    def __init__(self, encoding, encoding_errors, decode_responses: bool) -> None: ...
    def encode(self, value: str | bytes | memoryview | bool | float) -> bytes: ...
    def decode(self, value: str | bytes | memoryview, force: bool = ...) -> str: ...

class Connection:
    description_format: Any
    pid: Any
    host: Any
    port: Any
    db: Any
    password: Any
    socket_timeout: Any
    socket_connect_timeout: Any
    socket_keepalive: Any
    socket_keepalive_options: Any
    retry_on_timeout: Any
    retry_on_error: Any
    encoding: Any
    encoding_errors: Any
    decode_responses: Any
    retry: Retry
    redis_connect_func: Any | None
    def __init__(
        self,
        host: str = ...,
        port: int = ...,
        db: int = ...,
        password: str | None = ...,
        socket_timeout: float | None = ...,
        socket_connect_timeout: float | None = ...,
        socket_keepalive: bool = ...,
        socket_keepalive_options: Mapping[str, int | str] | None = ...,
        socket_type: int = ...,
        retry_on_timeout: bool = ...,
        retry_on_error=...,
        encoding: str = ...,
        encoding_errors: str = ...,
        decode_responses: bool = ...,
        parser_class: type[BaseParser] = ...,
        socket_read_size: int = ...,
        health_check_interval: int = ...,
        client_name: str | None = ...,
        username: str | None = ...,
        retry: Retry | None = ...,
        redis_connect_func: Any | None = ...,
    ) -> None: ...
    def __del__(self): ...
    def register_connect_callback(self, callback): ...
    def clear_connect_callbacks(self): ...
    def set_parser(self, parser_class): ...
    def connect(self): ...
    def on_connect(self): ...
    def disconnect(self, *args: object) -> None: ...  # 'args' added in redis 4.1.2
    def check_health(self) -> None: ...
    def send_packed_command(self, command, check_health: bool = ...): ...
    def send_command(self, *args): ...
    def can_read(self, timeout=...): ...
    def read_response(self, disable_decoding: bool = ...): ...
    def pack_command(self, *args): ...
    def pack_commands(self, commands): ...
    def repr_pieces(self) -> list[tuple[str, str]]: ...

class SSLConnection(Connection):
    description_format: Any
    keyfile: Any
    certfile: Any
    cert_reqs: Any
    ca_certs: Any
    ca_path: Any | None
    check_hostname: bool
    certificate_password: Any | None
    ssl_validate_ocsp: bool
    ssl_validate_ocsp_stapled: bool  # added in 4.1.1
    ssl_ocsp_context: Any | None  # added in 4.1.1
    ssl_ocsp_expected_cert: Any | None  # added in 4.1.1
    def __init__(
        self,
        ssl_keyfile=...,
        ssl_certfile=...,
        ssl_cert_reqs=...,
        ssl_ca_certs=...,
        ssl_check_hostname: bool = ...,
        ssl_ca_path: Any | None = ...,
        ssl_password: Any | None = ...,
        ssl_validate_ocsp: bool = ...,
        ssl_validate_ocsp_stapled: bool = ...,  # added in 4.1.1
        ssl_ocsp_context: Any | None = ...,  # added in 4.1.1
        ssl_ocsp_expected_cert: Any | None = ...,  # added in 4.1.1
        **kwargs,
    ) -> None: ...

class UnixDomainSocketConnection(Connection):
    description_format: Any
    pid: Any
    path: Any
    db: Any
    password: Any
    socket_timeout: Any
    retry_on_timeout: Any
    encoding: Any
    encoding_errors: Any
    decode_responses: Any
    retry: Retry
    def __init__(
        self,
        path=...,
        db: int = ...,
        username=...,
        password=...,
        socket_timeout=...,
        encoding: str = ...,
        encoding_errors: str = ...,
        decode_responses: bool = ...,
        retry_on_timeout: bool = ...,
        retry_on_error=...,
        parser_class=...,
        socket_read_size: int = ...,
        health_check_interval: int = ...,
        client_name=...,
        retry: Retry | None = ...,
        redis_connect_func: Any | None = ...,
    ) -> None: ...
    def repr_pieces(self) -> list[tuple[str, str]]: ...

class ConnectionPool:
    @classmethod
    def from_url(cls: type[Self], url: str, *, db: int = ..., decode_components: bool = ..., **kwargs) -> Self: ...
    connection_class: Any
    connection_kwargs: Any
    max_connections: Any
    def __init__(self, connection_class=..., max_connections=..., **connection_kwargs) -> None: ...
    pid: Any
    def reset(self): ...
    def get_connection(self, command_name, *keys, **options: _ConnectionPoolOptions): ...
    def make_connection(self): ...
    def release(self, connection): ...
    def disconnect(self, inuse_connections: bool = ...): ...
    def get_encoder(self) -> Encoder: ...
    def owns_connection(self, connection: Connection) -> bool: ...

class BlockingConnectionPool(ConnectionPool):
    queue_class: Any
    timeout: Any
    def __init__(self, max_connections=..., timeout=..., connection_class=..., queue_class=..., **connection_kwargs) -> None: ...
    pid: Any
    pool: Any
    def reset(self): ...
    def make_connection(self): ...
    def get_connection(self, command_name, *keys, **options: _ConnectionPoolOptions): ...
    def release(self, connection): ...
    def disconnect(self): ...

def to_bool(value: object) -> bool: ...
def parse_url(url: str) -> dict[str, Any]: ...



import sys
from typing import Callable

if sys.platform != "win32":
    from _curses import _CursesWindow
    def rectangle(win: _CursesWindow, uly: int, ulx: int, lry: int, lrx: int) -> None: ...

    class Textbox:
        stripspaces: bool
        def __init__(self, win: _CursesWindow, insert_mode: bool = ...) -> None: ...
        def edit(self, validate: Callable[[int], int] | None = ...) -> str: ...
        def do_command(self, ch: str | int) -> None: ...
        def gather(self) -> str: ...



from typing import Any, Iterable, TypeVar

__all__ = ["NodeList", "EmptyNodeList", "StringTypes", "defproperty"]

_T = TypeVar("_T")

StringTypes: tuple[type[str]]

class NodeList(list[_T]):
    length: int
    def item(self, index: int) -> _T | None: ...

class EmptyNodeList(tuple[Any, ...]):
    length: int
    def item(self, index: int) -> None: ...
    def __add__(self, other: Iterable[_T]) -> NodeList[_T]: ...  # type: ignore[override]
    def __radd__(self, other: Iterable[_T]) -> NodeList[_T]: ...

def defproperty(klass: type[Any], name: str, doc: str) -> None: ...



from typing import Any

from braintree.attribute_getter import AttributeGetter as AttributeGetter
from braintree.configuration import Configuration as Configuration
from braintree.resource import Resource as Resource
from braintree.risk_data import RiskData as RiskData
from braintree.three_d_secure_info import ThreeDSecureInfo as ThreeDSecureInfo

class CreditCardVerification(AttributeGetter):
    class Status:
        Failed: str
        GatewayRejected: str
        ProcessorDeclined: str
        Verified: str
    amount: Any
    currency_iso_code: Any
    processor_response_code: Any
    processor_response_text: Any
    network_response_code: Any
    network_response_text: Any
    risk_data: Any
    three_d_secure_info: Any
    def __init__(self, gateway, attributes) -> None: ...
    @staticmethod
    def find(verification_id): ...
    @staticmethod
    def search(*query): ...
    @staticmethod
    def create(params): ...
    @staticmethod
    def create_signature(): ...
    def __eq__(self, other): ...



from typing import Awaitable, Callable
from typing_extensions import ParamSpec

from .case import TestCase

_P = ParamSpec("_P")

class IsolatedAsyncioTestCase(TestCase):
    async def asyncSetUp(self) -> None: ...
    async def asyncTearDown(self) -> None: ...
    def addAsyncCleanup(self, __func: Callable[_P, Awaitable[object]], *args: _P.args, **kwargs: _P.kwargs) -> None: ...



import sys
from _typeshed import Self
from typing import Any, Generic, Iterable, TypeVar

if sys.version_info >= (3, 9):
    from types import GenericAlias

__all__ = ["SharedMemory", "ShareableList"]

_SLT = TypeVar("_SLT", int, float, bool, str, bytes, None)

class SharedMemory:
    def __init__(self, name: str | None = ..., create: bool = ..., size: int = ...) -> None: ...
    @property
    def buf(self) -> memoryview: ...
    @property
    def name(self) -> str: ...
    @property
    def size(self) -> int: ...
    def close(self) -> None: ...
    def unlink(self) -> None: ...

class ShareableList(Generic[_SLT]):
    shm: SharedMemory
    def __init__(self, sequence: Iterable[_SLT] | None = ..., *, name: str | None = ...) -> None: ...
    def __getitem__(self, position: int) -> _SLT: ...
    def __setitem__(self, position: int, value: _SLT) -> None: ...
    def __reduce__(self: Self) -> tuple[Self, tuple[_SLT, ...]]: ...
    def __len__(self) -> int: ...
    @property
    def format(self) -> str: ...
    def count(self, value: _SLT) -> int: ...
    def index(self, value: _SLT) -> int: ...
    if sys.version_info >= (3, 9):
        def __class_getitem__(cls, item: Any) -> GenericAlias: ...



from typing import IO, Any, Callable, Sequence

from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurve, EllipticCurvePrivateKey, EllipticCurvePublicKey
from cryptography.hazmat.primitives.hashes import HashAlgorithm
from paramiko.message import Message
from paramiko.pkey import PKey

class _ECDSACurve:
    nist_name: str
    key_length: int
    key_format_identifier: str
    hash_object: type[HashAlgorithm]
    curve_class: type[EllipticCurve]
    def __init__(self, curve_class: type[EllipticCurve], nist_name: str) -> None: ...

class _ECDSACurveSet:
    ecdsa_curves: Sequence[_ECDSACurve]
    def __init__(self, ecdsa_curves: Sequence[_ECDSACurve]) -> None: ...
    def get_key_format_identifier_list(self) -> list[str]: ...
    def get_by_curve_class(self, curve_class: type[Any]) -> _ECDSACurve | None: ...
    def get_by_key_format_identifier(self, key_format_identifier: str) -> _ECDSACurve | None: ...
    def get_by_key_length(self, key_length: int) -> _ECDSACurve | None: ...

class ECDSAKey(PKey):
    verifying_key: EllipticCurvePublicKey
    signing_key: EllipticCurvePrivateKey
    public_blob: None
    ecdsa_curve: _ECDSACurve | None
    def __init__(
        self,
        msg: Message | None = ...,
        data: bytes | None = ...,
        filename: str | None = ...,
        password: str | None = ...,
        vals: tuple[EllipticCurvePrivateKey, EllipticCurvePublicKey] | None = ...,
        file_obj: IO[str] | None = ...,
        validate_point: bool = ...,
    ) -> None: ...
    @classmethod
    def supported_key_format_identifiers(cls: Any) -> list[str]: ...
    def asbytes(self) -> bytes: ...
    def __hash__(self) -> int: ...
    def get_name(self) -> str: ...
    def get_bits(self) -> int: ...
    def can_sign(self) -> bool: ...
    def sign_ssh_data(self, data: bytes) -> Message: ...
    def verify_ssh_sig(self, data: bytes, msg: Message) -> bool: ...
    def write_private_key_file(self, filename: str, password: str | None = ...) -> None: ...
    def write_private_key(self, file_obj: IO[str], password: str | None = ...) -> None: ...
    @classmethod
    def generate(
        cls, curve: EllipticCurve = ..., progress_func: Callable[..., Any] | None = ..., bits: int | None = ...
    ) -> ECDSAKey: ...



from .scope import Scope
from .span import Span

class ScopeManager:
    def __init__(self) -> None: ...
    def activate(self, span: Span, finish_on_close: bool) -> Scope: ...
    @property
    def active(self) -> Scope | None: ...



import itertools

from typing import Any

import attrs

from pytype.pytd import visitors
from pytype.pytd.parse import node
import unittest


Node = node.Node


@attrs.frozen
class Node1(Node):
  """Simple node for equality testing. Not equal to anything else."""
  a: Any
  b: Any


@attrs.frozen
class Node2(Node):
  """For equality testing. Same attributes as Node3."""
  x: Any
  y: Any


@attrs.frozen
class Node3(Node):
  """For equality testing. Same attributes as Node2."""
  x: Any
  y: Any


@attrs.frozen
class Data(Node):
  """'Data' node. Visitor tests use this to store numbers in leafs."""
  d1: Any
  d2: Any
  d3: Any


@attrs.frozen
class V(Node):
  """Inner node 'V', with one child. See testVisitor[...]() below."""
  x: Any


@attrs.frozen
class X(Node):
  """Inner node 'X', with two children. See testVisitor[...]() below."""
  a: Any
  b: Any


@attrs.frozen
class Y(Node):
  """Inner node 'Y', with two children. See testVisitor[...]() below."""
  c: Any
  d: Any


@attrs.frozen
class XY(Node):
  """Inner node 'XY', with two children. See testVisitor[...]() below."""
  x: Any
  y: Any


@attrs.frozen
class NodeWithVisit(Node):
  """A node with its own VisitNode function."""
  x: Any
  y: Any

  def VisitNode(self, visitor):
    """Allow a visitor to modify our children. Returns modified node."""
    # only visit x, not y
    x = self.x.Visit(visitor)
    return NodeWithVisit(x, self.y)


class DataVisitor(visitors.Visitor):
  """A visitor that transforms Data nodes."""

  def VisitData(self, data):
    """Visit Data nodes, and set 'd3' attribute to -1."""
    return data.Replace(d3=-1)


class MultiNodeVisitor(visitors.Visitor):
  """A visitor that visits Data, V and Y nodes and uses the *args feature."""

  def VisitData(self, _, r):
    """Visit Data nodes, change them to XY nodes, and set x and y."""
    return XY(r, r)

  def VisitV(self, _, r):
    """Visit V nodes, change them to X nodes with V nodes as children."""
    return X(V(r), V(r))

  def VisitY(self, y):
    """Visit Y nodes, and change them to X nodes with the same attributes."""
    return X(*y)


class SkipNodeVisitor(visitors.Visitor):
  """A visitor that skips XY.y subtrees."""

  def EnterXY(self, _):
    return {"y"}

  def VisitData(self, data):
    """Visit Data nodes, and zero all data."""
    return data.Replace(d1=0, d2=0, d3=0)


# We want to test == and != so:
# pylint: disable=g-generic-assert
class TestNode(unittest.TestCase):
  """Test the node.Node class generator."""

  def test_eq1(self):
    """Test the __eq__ and __ne__ functions of node.Node."""
    n1 = Node1(a=1, b=2)
    n2 = Node1(a=1, b=2)
    self.assertEqual(n1, n2)
    self.assertFalse(n1 != n2)

  def test_hash1(self):
    n1 = Node1(a=1, b=2)
    n2 = Node1(a=1, b=2)
    self.assertEqual(hash(n1), hash(n2))

  def test_eq2(self):
    """Test the __eq__ and __ne__ functions of identical nested nodes."""
    n1 = Node1(a=1, b=2)
    n2 = Node1(a=1, b=2)
    d1 = Node2(x="foo", y=n1)
    d2 = Node2(x="foo", y=n1)
    d3 = Node2(x="foo", y=n2)
    d4 = Node2(x="foo", y=n2)
    self.assertTrue(d1 == d2 and d2 == d3 and d3 == d4 and d4 == d1)
    # Since node overloads __ne___, too, test it explicitly:
    self.assertFalse(d1 != d2 or d2 != d3 or d3 != d4 or d4 != d1)

  def test_hash2(self):
    n1 = Node1(a=1, b=2)
    n2 = Node1(a=1, b=2)
    d1 = Node2(x="foo", y=n1)
    d2 = Node2(x="foo", y=n1)
    d3 = Node2(x="foo", y=n2)
    d4 = Node2(x="foo", y=n2)
    self.assertEqual(hash(d1), hash(d2))
    self.assertEqual(hash(d2), hash(d3))
    self.assertEqual(hash(d3), hash(d4))
    self.assertEqual(hash(d4), hash(d1))

  def test_deep_eq2(self):
    """Test the __eq__ and __ne__ functions of differing nested nodes."""
    n1 = Node1(a=1, b=2)
    n2 = Node1(a=1, b=3)
    d1 = Node2(x="foo", y=n1)
    d2 = Node3(x="foo", y=n1)
    d3 = Node2(x="foo", y=n2)
    d4 = Node3(x="foo", y=n2)
    self.assertNotEqual(d1, d2)
    self.assertNotEqual(d1, d3)
    self.assertNotEqual(d1, d4)
    self.assertNotEqual(d2, d3)
    self.assertNotEqual(d2, d4)
    self.assertNotEqual(d3, d4)
    self.assertFalse(d1 == d2)
    self.assertFalse(d1 == d3)
    self.assertFalse(d1 == d4)
    self.assertFalse(d2 == d3)
    self.assertFalse(d2 == d4)
    self.assertFalse(d3 == d4)

  def test_deep_hash2(self):
    n1 = Node1(a=1, b=2)
    n2 = Node1(a=1, b=3)
    d1 = Node2(x="foo", y=n1)
    d2 = Node3(x="foo", y=n1)
    d3 = Node2(x="foo", y=n2)
    d4 = Node3(x="foo", y=n2)
    self.assertNotEqual(hash(d1), hash(d2))
    self.assertNotEqual(hash(d1), hash(d3))
    self.assertNotEqual(hash(d1), hash(d4))
    self.assertNotEqual(hash(d2), hash(d3))
    self.assertNotEqual(hash(d2), hash(d4))
    self.assertNotEqual(hash(d3), hash(d4))

  def test_immutable(self):
    """Test that node.Node has/preserves immutatibility."""
    n1 = Node1(a=1, b=2)
    n2 = Node2(x="foo", y=n1)
    with self.assertRaises(AttributeError):
      n1.a = 2
    with self.assertRaises(AttributeError):
      n2.x = "bar"
    with self.assertRaises(AttributeError):
      n2.x.b = 3

  def test_visitor1(self):
    """Test node.Node.Visit() for a visitor that modifies leaf nodes."""
    x = X(1, (1, 2))
    y = Y((V(1),), Data(42, 43, 44))
    xy = XY(x, y)
    xy_expected = ("XY(x=X(a=1, b=(1, 2)), y=Y(c=(V(x=1),),"
                   " d=Data(d1=42, d2=43, d3=44)))")
    self.assertEqual(repr(xy), xy_expected)
    v = DataVisitor()
    new_xy = xy.Visit(v)
    self.assertEqual(repr(new_xy),
                     "XY(x=X(a=1, b=(1, 2)), y=Y(c=(V(x=1),),"
                     " d=Data(d1=42, d2=43, d3=-1)))")
    self.assertEqual(repr(xy), xy_expected)  # check that xy is unchanged

  def test_visitor2(self):
    """Test node.Node.Visit() for visitors that modify inner nodes."""
    xy = XY(V(1), Data(1, 2, 3))
    xy_expected = "XY(x=V(x=1), y=Data(d1=1, d2=2, d3=3))"
    self.assertEqual(repr(xy), xy_expected)
    v = MultiNodeVisitor()
    new_xy = xy.Visit(v, 42)
    self.assertEqual(repr(new_xy),
                     "XY(x=X(a=V(x=42), b=V(x=42)), y=XY(x=42, y=42))")
    self.assertEqual(repr(xy), xy_expected)  # check that xy is unchanged

  def test_skip_visitor(self):
    tree = XY(V(Data(1, 2, 3)), XY(Data(3, 4, 5), Data(6, 7, 8)))
    init = ("XY(x=V(x=Data(d1=1, d2=2, d3=3)), y=XY(x=Data(d1=3, d2=4, d3=5), "
            "y=Data(d1=6, d2=7, d3=8)))")
    self.assertEqual(repr(tree), init)
    new_tree = tree.Visit(SkipNodeVisitor())
    exp = ("XY(x=V(x=Data(d1=0, d2=0, d3=0)), y=XY(x=Data(d1=3, d2=4, d3=5), "
           "y=Data(d1=6, d2=7, d3=8)))")
    self.assertEqual(repr(new_tree), exp)

  def test_recursion(self):
    """Test node.Node.Visit() for visitors that preserve attributes."""
    y = Y(Y(1, 2), Y(3, Y(4, 5)))
    y_expected = "Y(c=Y(c=1, d=2), d=Y(c=3, d=Y(c=4, d=5)))"
    self.assertEqual(repr(y), y_expected)
    v = MultiNodeVisitor()
    new_y = y.Visit(v)
    new_repr = "X(a=X(a=1, b=2), b=X(a=3, b=X(a=4, b=5)))"
    self.assertEqual(repr(new_y), new_repr)
    self.assertEqual(repr(y), y_expected)  # check that original is unchanged

  def test_tuple(self):
    """Test node.Node.Visit() for nodes that contain tuples."""
    v = V((Data(1, 2, 3), Data(4, 5, 6)))
    v_expected = "V(x=(Data(d1=1, d2=2, d3=3), Data(d1=4, d2=5, d3=6)))"
    self.assertEqual(repr(v), v_expected)
    visit = DataVisitor()
    new_v = v.Visit(visit)
    new_v_expected = "V(x=(Data(d1=1, d2=2, d3=-1), Data(d1=4, d2=5, d3=-1)))"
    self.assertEqual(repr(new_v), new_v_expected)

  def test_ordering(self):
    nodes = [Node1(True, False), Node1(1, 2),
             Node2(1, 1), Node2("2", "1"),
             Node3(1, 1), Node3(2, 2),
             V(2)]
    for n1, n2 in zip(nodes[:-1], nodes[1:]):
      self.assertLess(n1, n2)
      self.assertLessEqual(n1, n2)
      self.assertGreater(n2, n1)
      self.assertGreaterEqual(n2, n1)
    for p in itertools.permutations(nodes):
      self.assertEqual(list(sorted(p)), nodes)

# pylint: enable=g-generic-assert


if __name__ == "__main__":
  unittest.main()



"""Tests for pytype_runner.py."""

import collections
import dataclasses
import os

from pytype import config as pytype_config
from pytype import file_utils
from pytype import module_utils
from pytype.tools.analyze_project import parse_args
from pytype.tools.analyze_project import pytype_runner
import unittest


# Convenience aliases.
# pylint: disable=invalid-name
Module = module_utils.Module
Action = pytype_runner.Action
Stage = pytype_runner.Stage
# pylint: enable=invalid-name


# named 'Local' to match importlab.resolve.Local
@dataclasses.dataclass(eq=True, frozen=True)
class Local:
  path: str
  short_path: str
  module_name: str


@dataclasses.dataclass(eq=True, frozen=True)
class ExpectedBuildStatement:
  output: str
  action: str
  input: str
  deps: str
  imports: str
  module: str


# number of lines in the build.ninja preamble
_PREAMBLE_LENGTH = 6


class FakeImportGraph:
  """Just enough of the ImportGraph interface to run tests."""

  def __init__(self, source_files, provenance, source_to_deps):
    self.source_files = source_files
    self.provenance = provenance
    self.source_to_deps = source_to_deps

  def deps_list(self):
    return [(x, self.source_to_deps[x]) for x in reversed(self.source_files)]


def make_runner(sources, dep, conf):
  conf.inputs = [m.full_path for m in sources]
  return pytype_runner.PytypeRunner(conf, dep)


class TestResolvedFileToModule(unittest.TestCase):
  """Test resolved_file_to_module."""

  def test_basic(self):
    resolved_file = Local('foo/bar.py', 'bar.py', 'bar')
    self.assertEqual(pytype_runner.resolved_file_to_module(resolved_file),
                     Module('foo/', 'bar.py', 'bar', 'Local'))

  def test_preserve_init(self):
    resolved_file = Local('foo/bar/__init__.py', 'bar/__init__.py', 'bar')
    self.assertEqual(pytype_runner.resolved_file_to_module(resolved_file),
                     Module('foo/', 'bar/__init__.py', 'bar.__init__', 'Local'))


class TestDepsFromImportGraph(unittest.TestCase):
  """Test deps_from_import_graph."""

  def setUp(self):
    super().setUp()
    init = Local('/foo/bar/__init__.py', 'bar/__init__.py', 'bar')
    a = Local('/foo/bar/a.py', 'bar/a.py', 'bar.a')
    b = Local('/foo/bar/b.py', 'bar/b.py', 'bar.b')
    self.sources = [x.path for x in [init, a, b]]
    self.provenance = {x.path: x for x in [init, a, b]}

  def test_basic(self):
    graph = FakeImportGraph(self.sources, self.provenance,
                            collections.defaultdict(list))
    deps = pytype_runner.deps_from_import_graph(graph)
    expected = [
        ((Module('/foo/', 'bar/__init__.py', 'bar.__init__'),), ()),
        ((Module('/foo/', 'bar/a.py', 'bar.a'),), ()),
        ((Module('/foo/', 'bar/b.py', 'bar.b'),), ()),
    ]
    self.assertEqual(deps, expected)

  def test_duplicate_deps(self):
    graph = FakeImportGraph(
        self.sources, self.provenance,
        collections.defaultdict(lambda: [self.sources[0]] * 2))
    deps = pytype_runner.deps_from_import_graph(graph)
    init = Module('/foo/', 'bar/__init__.py', 'bar.__init__')
    expected = [
        ((init,), (init,)),
        ((Module('/foo/', 'bar/a.py', 'bar.a'),), (init,)),
        ((Module('/foo/', 'bar/b.py', 'bar.b'),), (init,)),
    ]
    self.assertEqual(deps, expected)

  def test_pyi_src(self):
    pyi_mod = Local('/foo/bar/c.pyi', 'bar/c.pyi', 'bar.c')
    provenance = {pyi_mod.path: pyi_mod}
    provenance.update(self.provenance)
    graph = FakeImportGraph(self.sources + [pyi_mod.path], provenance,
                            collections.defaultdict(list))
    deps = pytype_runner.deps_from_import_graph(graph)
    expected = [
        ((Module('/foo/', 'bar/__init__.py', 'bar.__init__'),), ()),
        ((Module('/foo/', 'bar/a.py', 'bar.a'),), ()),
        ((Module('/foo/', 'bar/b.py', 'bar.b'),), ()),
    ]
    self.assertEqual(deps, expected)

  def test_pyi_dep(self):
    pyi_mod = Local('/foo/bar/c.pyi', 'bar/c.pyi', 'bar.c')
    graph = FakeImportGraph(self.sources, self.provenance,
                            collections.defaultdict(lambda: [pyi_mod.path]))
    deps = pytype_runner.deps_from_import_graph(graph)
    expected = [
        ((Module('/foo/', 'bar/__init__.py', 'bar.__init__'),), ()),
        ((Module('/foo/', 'bar/a.py', 'bar.a'),), ()),
        ((Module('/foo/', 'bar/b.py', 'bar.b'),), ()),
    ]
    self.assertEqual(deps, expected)


class TestBase(unittest.TestCase):
  """Base class for tests using a parser."""

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.parser = parse_args.make_parser()


class TestCustomOptions(TestBase):
  """Test PytypeRunner.set_custom_options."""

  def setUp(self):
    super().setUp()
    self.conf = self.parser.config_from_defaults()

  def assertFlags(self, flags, expected_flags):
    # Add temporary flags that are set to true by default here, so that they are
    # filtered out of tests.
    temporary_flags = set()
    self.assertEqual(flags - temporary_flags, expected_flags)

  # --disable tests a flag with a string value.

  def test_disable(self):
    self.conf.disable = ['import-error', 'name-error']
    runner = make_runner([], [], self.conf)
    flags_with_values = {}
    runner.set_custom_options(flags_with_values, set(), self.conf.report_errors)
    self.assertEqual(flags_with_values['--disable'], 'import-error,name-error')

  def test_no_disable(self):
    self.conf.disable = []
    runner = make_runner([], [], self.conf)
    flags_with_values = {}
    runner.set_custom_options(flags_with_values, set(), self.conf.report_errors)
    self.assertFalse(flags_with_values)

  # The purpose of the following --no-report-errors tests is to test a generic
  # binary flag with a custom to_command_line. These tests do not reflect actual
  # error-reporting behavior; for that, see TestGetRunCmd.test_error_reporting.

  def test_report_errors(self):
    self.conf.report_errors = True
    runner = make_runner([], [], self.conf)
    binary_flags = {'--no-report-errors'}
    runner.set_custom_options({}, binary_flags, True)
    self.assertFlags(binary_flags, set())

  def test_no_report_errors(self):
    self.conf.report_errors = False
    runner = make_runner([], [], self.conf)
    binary_flags = set()
    runner.set_custom_options({}, binary_flags, True)
    self.assertFlags(binary_flags, {'--no-report-errors'})

  def test_report_errors_default(self):
    self.conf.report_errors = True
    runner = make_runner([], [], self.conf)
    binary_flags = set()
    runner.set_custom_options({}, binary_flags, True)
    self.assertFlags(binary_flags, set())

  # --protocols tests a binary flag whose value is passed through transparently.

  def test_protocols(self):
    self.conf.protocols = True
    runner = make_runner([], [], self.conf)
    binary_flags = set()
    runner.set_custom_options({}, binary_flags, self.conf.report_errors)
    self.assertFlags(binary_flags, {'--protocols'})

  def test_no_protocols(self):
    self.conf.protocols = False
    runner = make_runner([], [], self.conf)
    binary_flags = {'--protocols'}
    runner.set_custom_options({}, binary_flags, self.conf.report_errors)
    self.assertFlags(binary_flags, set())

  def test_no_protocols_default(self):
    self.conf.protocols = False
    runner = make_runner([], [], self.conf)
    binary_flags = set()
    runner.set_custom_options({}, binary_flags, self.conf.report_errors)
    self.assertFlags(binary_flags, set())


class TestGetRunCmd(TestBase):
  """Test PytypeRunner.get_pytype_command_for_ninja()."""

  def setUp(self):
    super().setUp()
    self.runner = make_runner([], [], self.parser.config_from_defaults())

  def get_options(self, args):
    nargs = len(pytype_runner.PYTYPE_SINGLE)
    self.assertEqual(args[:nargs], pytype_runner.PYTYPE_SINGLE)
    args = args[nargs:]
    start, end = args.index('--imports_info'), args.index('$imports')
    self.assertEqual(end-start, 1)
    args.pop(end)
    args.pop(start)
    return pytype_config.Options(args, command_line=True)

  def get_basic_options(self, report_errors=False):
    return self.get_options(
        self.runner.get_pytype_command_for_ninja(report_errors))

  def test_python_version(self):
    self.assertEqual(
        self.get_basic_options().python_version,
        tuple(int(i) for i in self.runner.python_version.split('.')))

  def test_output(self):
    self.assertEqual(self.get_basic_options().output, '$out')

  def test_quick(self):
    self.assertTrue(self.get_basic_options().quick)

  def test_module_name(self):
    self.assertEqual(self.get_basic_options().module_name, '$module')

  def test_error_reporting(self):
    # Disable error reporting
    options = self.get_basic_options(report_errors=False)
    self.assertFalse(options.report_errors)
    self.assertFalse(options.analyze_annotated)
    # Enable error reporting
    options = self.get_basic_options(report_errors=True)
    self.assertTrue(options.report_errors)
    self.assertTrue(options.analyze_annotated)

  def test_custom_option(self):
    custom_conf = self.parser.config_from_defaults()
    custom_conf.disable = ['import-error', 'name-error']
    self.runner = make_runner([], [], custom_conf)
    args = self.runner.get_pytype_command_for_ninja(report_errors=True)
    options = self.get_options(args)
    self.assertEqual(options.disable, ['import-error', 'name-error'])

  def test_custom_option_no_report_errors(self):
    custom_conf = self.parser.config_from_defaults()
    # If the --precise-return flag is ever removed, replace it with another
    # feature or experimental flag from pytype.config - preferably one expected
    # to be long-lived to reduce churn.
    custom_conf.precise_return = True
    self.runner = make_runner([], [], custom_conf)
    args = self.runner.get_pytype_command_for_ninja(report_errors=False)
    options = self.get_options(args)
    self.assertTrue(options.precise_return)


class TestGetModuleAction(TestBase):
  """Tests for PytypeRunner.get_module_action."""

  def test_check(self):
    sources = [Module('', 'foo.py', 'foo')]
    runner = make_runner(sources, [], self.parser.config_from_defaults())
    self.assertEqual(runner.get_module_action(sources[0]),
                     pytype_runner.Action.CHECK)

  def test_infer(self):
    runner = make_runner([], [], self.parser.config_from_defaults())
    self.assertEqual(runner.get_module_action(Module('', 'foo.py', 'foo')),
                     pytype_runner.Action.INFER)

  def test_generate_default(self):
    runner = make_runner([], [], self.parser.config_from_defaults())
    self.assertEqual(
        runner.get_module_action(Module('', 'foo.py', 'foo', 'System')),
        pytype_runner.Action.GENERATE_DEFAULT)


class TestYieldSortedModules(TestBase):
  """Tests for PytypeRunner.yield_sorted_modules()."""

  def normalize(self, d):
    return file_utils.expand_path(d).rstrip(os.sep) + os.sep

  def assert_sorted_modules_equal(self, mod_gen, expected_list):
    for (expected_mod,
         expected_report_errors,
         expected_deps,
         expected_stage) in expected_list:
      try:
        mod, actual_report_errors, actual_deps, actual_stage = next(mod_gen)
      except StopIteration as e:
        raise AssertionError('Not enough modules') from e
      self.assertEqual(mod, expected_mod)
      self.assertEqual(actual_report_errors, expected_report_errors)
      self.assertEqual(actual_deps, expected_deps)
      self.assertEqual(actual_stage, expected_stage)
    try:
      next(mod_gen)
    except StopIteration:
      pass
    else:
      # Too many modules
      raise AssertionError('Too many modules')

  def test_source(self):
    conf = self.parser.config_from_defaults()
    d = self.normalize('foo/')
    conf.pythonpath = [d]
    f = Module(d, 'bar.py', 'bar')
    runner = make_runner([f], [((f,), ())], conf)
    self.assert_sorted_modules_equal(
        runner.yield_sorted_modules(),
        [(f, Action.CHECK, (), Stage.SINGLE_PASS)])

  def test_source_and_dep(self):
    conf = self.parser.config_from_defaults()
    d = self.normalize('foo/')
    conf.pythonpath = [d]
    src = Module(d, 'bar.py', 'bar')
    dep = Module(d, 'baz.py', 'baz')
    runner = make_runner([src], [((dep,), ()), ((src,), (dep,))], conf)
    self.assert_sorted_modules_equal(
        runner.yield_sorted_modules(),
        [(dep, Action.INFER, (), Stage.SINGLE_PASS),
         (src, Action.CHECK, (dep,), Stage.SINGLE_PASS)])

  def test_cycle(self):
    conf = self.parser.config_from_defaults()
    d = self.normalize('foo/')
    conf.pythonpath = [d]
    src = Module(d, 'bar.py', 'bar')
    dep = Module(d, 'baz.py', 'baz')
    runner = make_runner([src], [((dep, src), ())], conf)
    self.assert_sorted_modules_equal(
        runner.yield_sorted_modules(),
        [(dep, Action.INFER, (), Stage.FIRST_PASS),
         (src, Action.INFER, (), Stage.FIRST_PASS),
         (dep, Action.INFER, (dep, src), Stage.SECOND_PASS),
         (src, Action.CHECK, (dep, src), Stage.SECOND_PASS)])

  def test_system_dep(self):
    conf = self.parser.config_from_defaults()
    d = self.normalize('foo/')
    external = self.normalize('quux/')
    conf.pythonpath = [d]
    mod = Module(external, 'bar/baz.py', 'bar.baz', 'System')
    runner = make_runner([], [((mod,), ())], conf)
    self.assert_sorted_modules_equal(
        runner.yield_sorted_modules(),
        [(mod, Action.GENERATE_DEFAULT, (), Stage.SINGLE_PASS)])


class TestNinjaPreamble(TestBase):
  """Tests for PytypeRunner.write_ninja_preamble."""

  def test_write(self):
    conf = self.parser.config_from_defaults()
    with file_utils.Tempdir() as d:
      conf.output = d.path
      runner = make_runner([], [], conf)
      runner.write_ninja_preamble()
      with open(runner.ninja_file) as f:
        preamble = f.read().splitlines()
    self.assertEqual(len(preamble), _PREAMBLE_LENGTH)
    # The preamble consists of triples of lines of the format:
    # rule {name}
    #   command = pytype-single {args} $in
    #   description = {name} $module
    # Check that the lines cycle through these patterns.
    for i, line in enumerate(preamble):
      if not i % 3:
        self.assertRegex(line, r'rule \w*')
      elif i % 3 == 1:
        expected = r'  command = {} .* \$in'.format(
            ' '.join(pytype_runner.PYTYPE_SINGLE))
        self.assertRegex(line, expected)
      else:
        self.assertRegex(line, r'  description = \w* \$module')


class TestNinjaBuildStatement(TestBase):
  """Tests for PytypeRunner.write_build_statement."""

  def write_build_statement(self, *args, **kwargs):
    conf = self.parser.config_from_defaults()
    with file_utils.Tempdir() as d:
      conf.output = d.path
      runner = make_runner([], [], conf)
      output = runner.write_build_statement(*args, **kwargs)
      with open(runner.ninja_file) as f:
        return runner, output, f.read().splitlines()

  def assertOutputMatches(self, module, expected_output):
    runner, output, _ = self.write_build_statement(
        module, Action.CHECK, set(), 'imports', '')
    self.assertEqual(output, os.path.join(runner.pyi_dir, expected_output))

  def test_check(self):
    _, output, build_statement = self.write_build_statement(
        Module('', 'foo.py', 'foo'), Action.CHECK, set(), 'imports', '')
    self.assertEqual(build_statement[0], f'build {output}: check foo.py')

  def test_infer(self):
    _, output, build_statement = self.write_build_statement(
        Module('', 'foo.py', 'foo'), Action.INFER, set(), 'imports', '')
    self.assertEqual(build_statement[0], f'build {output}: infer foo.py')

  def test_deps(self):
    _, output, _ = self.write_build_statement(
        Module('', 'foo.py', 'foo'), Action.INFER, set(), 'imports', '')
    _, _, build_statement = self.write_build_statement(
        Module('', 'bar.py', 'bar'), Action.CHECK, {output}, 'imports', '')
    self.assertTrue(build_statement[0].endswith(' | ' + output))

  def test_imports(self):
    _, _, build_statement = self.write_build_statement(
        Module('', 'foo.py', 'foo'), Action.CHECK, set(), 'imports', '')
    self.assertIn('  imports = imports', build_statement)

  def test_module(self):
    _, _, build_statement = self.write_build_statement(
        Module('', 'foo.py', 'foo'), Action.CHECK, set(), 'imports', '')
    self.assertIn('  module = foo', build_statement)

  def test_suffix(self):
    runner, output, _ = self.write_build_statement(
        Module('', 'foo.py', 'foo'), Action.CHECK, set(), 'imports', '-1')
    self.assertEqual(os.path.join(runner.pyi_dir, 'foo.pyi-1'), output)

  def test_hidden_dir(self):
    self.assertOutputMatches(Module('', '.foo/bar.py', '.foo.bar'),
                             os.path.join('.foo', 'bar.pyi'))

  def test_hidden_file(self):
    self.assertOutputMatches(Module('', 'foo/.bar.py', 'foo..bar'),
                             os.path.join('foo', '.bar.pyi'))

  def test_hidden_file_with_path_prefix(self):
    self.assertOutputMatches(Module('', 'foo/.bar.py', '.bar'),
                             os.path.join('.bar.pyi'))

  def test_hidden_dir_with_path_mismatch(self):
    self.assertOutputMatches(Module('', 'symlinked/foo.py', '.bar'), '.bar.pyi')

  def test_path_mismatch(self):
    self.assertOutputMatches(Module('', 'symlinked/foo.py', 'bar.baz'),
                             os.path.join('bar', 'baz.pyi'))


class TestNinjaBody(TestBase):
  """Test PytypeRunner.setup_build."""

  def setUp(self):
    super().setUp()
    self.conf = self.parser.config_from_defaults()

  def assertBuildStatementMatches(self, build_statement, expected):
    self.assertEqual(build_statement[0],
                     'build {output}: {action} {input}{deps}'.format(
                         output=expected.output,
                         action=expected.action,
                         input=expected.input,
                         deps=expected.deps))
    self.assertEqual(set(build_statement[1:]),
                     {f'  imports = {expected.imports}',
                      f'  module = {expected.module}'})

  def test_basic(self):
    src = Module('', 'foo.py', 'foo')
    dep = Module('', 'bar.py', 'bar')
    with file_utils.Tempdir() as d:
      self.conf.output = d.path
      runner = make_runner(
          [src], [((dep,), ()), ((src,), (dep,))], self.conf)
      runner.setup_build()
      with open(runner.ninja_file) as f:
        body = f.read().splitlines()[_PREAMBLE_LENGTH:]
    self.assertBuildStatementMatches(body[0:3], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'bar.pyi'),
        action=Action.INFER,
        input='bar.py',
        deps='',
        imports=os.path.join(runner.imports_dir, 'bar.imports'),
        module='bar'))
    self.assertBuildStatementMatches(body[3:], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'foo.pyi'),
        action=Action.CHECK,
        input='foo.py',
        deps=' | ' + os.path.join(runner.pyi_dir, 'bar.pyi'),
        imports=os.path.join(runner.imports_dir, 'foo.imports'),
        module='foo'))

  def test_generate_default(self):
    src = Module('', 'foo.py', 'foo')
    dep = Module('', 'bar.py', 'bar', 'System')
    with file_utils.Tempdir() as d:
      self.conf.output = d.path
      runner = make_runner(
          [src], [((dep,), ()), ((src,), (dep,))], self.conf)
      runner.setup_build()
      with open(runner.ninja_file) as f:
        body = f.read().splitlines()[_PREAMBLE_LENGTH:]
      with open(os.path.join(runner.imports_dir, 'foo.imports')) as f:
        imports_info, = f.read().splitlines()
    self.assertBuildStatementMatches(body, ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'foo.pyi'),
        action=Action.CHECK,
        input='foo.py',
        deps='',
        imports=os.path.join(runner.imports_dir, 'foo.imports'),
        module='foo'))
    short_bar_path, bar_path = imports_info.split(' ')
    self.assertEqual(short_bar_path, 'bar')
    self.assertEqual(bar_path, os.path.join(runner.imports_dir, 'default.pyi'))

  def test_cycle(self):
    src = Module('', 'foo.py', 'foo')
    dep = Module('', 'bar.py', 'bar')
    with file_utils.Tempdir() as d:
      self.conf.output = d.path
      runner = make_runner(
          [src], [((dep, src), ())], self.conf)
      runner.setup_build()
      with open(runner.ninja_file) as f:
        body = f.read().splitlines()[_PREAMBLE_LENGTH:]
    self.assertBuildStatementMatches(body[:3], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'bar.pyi-1'),
        action=Action.INFER,
        input='bar.py',
        deps='',
        imports=os.path.join(runner.imports_dir, 'bar.imports-1'),
        module='bar'))
    self.assertBuildStatementMatches(body[3:6], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'foo.pyi-1'),
        action=Action.INFER,
        input='foo.py',
        deps='',
        imports=os.path.join(runner.imports_dir, 'foo.imports-1'),
        module='foo'))
    self.assertBuildStatementMatches(body[6:9], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'bar.pyi'),
        action=Action.INFER,
        input='bar.py',
        deps=' | {} {}'.format(os.path.join(runner.pyi_dir, 'bar.pyi-1'),
                               os.path.join(runner.pyi_dir, 'foo.pyi-1')),
        imports=os.path.join(runner.imports_dir, 'bar.imports'),
        module='bar'))
    self.assertBuildStatementMatches(body[9:], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'foo.pyi'),
        action=Action.CHECK,
        input='foo.py',
        deps=' | {} {}'.format(os.path.join(runner.pyi_dir, 'bar.pyi'),
                               os.path.join(runner.pyi_dir, 'foo.pyi-1')),
        imports=os.path.join(runner.imports_dir, 'foo.imports'),
        module='foo'))

  def test_cycle_with_extra_action(self):
    src = Module('', 'foo.py', 'foo')
    dep = Module('', 'bar.py', 'bar')
    with file_utils.Tempdir() as d:
      self.conf.output = d.path
      # When `src` is analyzed before `dep`, the second infer action on `dep`
      # should be skipped.
      runner = make_runner([src], [((src, dep), ())], self.conf)
      runner.setup_build()
      with open(runner.ninja_file) as f:
        body = f.read().splitlines()[_PREAMBLE_LENGTH:]
    self.assertBuildStatementMatches(body[:3], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'foo.pyi-1'),
        action=Action.INFER,
        input='foo.py',
        deps='',
        imports=os.path.join(runner.imports_dir, 'foo.imports-1'),
        module='foo'))
    self.assertBuildStatementMatches(body[3:6], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'bar.pyi-1'),
        action=Action.INFER,
        input='bar.py',
        deps='',
        imports=os.path.join(runner.imports_dir, 'bar.imports-1'),
        module='bar'))
    self.assertBuildStatementMatches(body[6:], ExpectedBuildStatement(
        output=os.path.join(runner.pyi_dir, 'foo.pyi'),
        action=Action.CHECK,
        input='foo.py',
        deps=' | {} {}'.format(os.path.join(runner.pyi_dir, 'foo.pyi-1'),
                               os.path.join(runner.pyi_dir, 'bar.pyi-1')),
        imports=os.path.join(runner.imports_dir, 'foo.imports'),
        module='foo'))


class TestImports(TestBase):
  """Test imports-related functionality."""

  def setUp(self):
    super().setUp()
    self.conf = self.parser.config_from_defaults()

  def test_write_default_pyi(self):
    with file_utils.Tempdir() as d:
      self.conf.output = d.path
      runner = make_runner([], [], self.conf)
      self.assertTrue(runner.make_imports_dir())
      output = runner.write_default_pyi()
      self.assertEqual(output, os.path.join(runner.imports_dir, 'default.pyi'))
      with open(output) as f:
        self.assertEqual(f.read(), pytype_runner.DEFAULT_PYI)

  def test_write_imports(self):
    with file_utils.Tempdir() as d:
      self.conf.output = d.path
      runner = make_runner([], [], self.conf)
      self.assertTrue(runner.make_imports_dir())
      output = runner.write_imports('mod', {'a': 'b'}, '')
      self.assertEqual(os.path.join(runner.imports_dir, 'mod.imports'), output)
      with open(output) as f:
        self.assertEqual(f.read(), 'a b\n')

  def test_get_imports_map(self):
    mod = Module('', 'foo.py', 'foo')
    deps = (mod,)
    module_to_imports_map = {mod: {'bar': '/dir/bar.pyi'}}
    module_to_output = {mod: '/dir/foo.pyi'}
    imports_map = pytype_runner.get_imports_map(
        deps, module_to_imports_map, module_to_output)
    self.assertEqual(imports_map,
                     {'foo': '/dir/foo.pyi', 'bar': '/dir/bar.pyi'})


if __name__ == '__main__':
  unittest.main()



from typing import IO, Any, overload

__all__ = [
    "get_config_h_filename",
    "get_config_var",
    "get_config_vars",
    "get_makefile_filename",
    "get_path",
    "get_path_names",
    "get_paths",
    "get_platform",
    "get_python_version",
    "get_scheme_names",
    "parse_config_h",
]

def get_config_var(name: str) -> str | None: ...
@overload
def get_config_vars() -> dict[str, Any]: ...
@overload
def get_config_vars(arg: str, *args: str) -> list[Any]: ...
def get_scheme_names() -> tuple[str, ...]: ...
def get_path_names() -> tuple[str, ...]: ...
def get_path(name: str, scheme: str = ..., vars: dict[str, Any] | None = ..., expand: bool = ...) -> str: ...
def get_paths(scheme: str = ..., vars: dict[str, Any] | None = ..., expand: bool = ...) -> dict[str, str]: ...
def get_python_version() -> str: ...
def get_platform() -> str: ...
def is_python_build(check_home: bool = ...) -> bool: ...
def parse_config_h(fp: IO[Any], vars: dict[str, Any] | None = ...) -> dict[str, Any]: ...
def get_config_h_filename() -> str: ...
def get_makefile_filename() -> str: ...



import json
from typing import Any

from stripe import api_requestor as api_requestor

class StripeObject(dict[Any, Any]):
    class ReprJSONEncoder(json.JSONEncoder):
        def default(self, obj): ...

    def __init__(
        self,
        id: Any | None = ...,
        api_key: Any | None = ...,
        stripe_version: Any | None = ...,
        stripe_account: Any | None = ...,
        last_response: Any | None = ...,
        **params,
    ) -> None: ...
    @property
    def last_response(self): ...
    def update(self, update_dict): ...
    def __setattr__(self, k, v): ...
    def __getattr__(self, k): ...
    def __delattr__(self, k): ...
    def __setitem__(self, k, v) -> None: ...
    def __getitem__(self, k): ...
    def __delitem__(self, k) -> None: ...
    def __reduce__(self): ...
    @classmethod
    def construct_from(
        cls, values, key, stripe_version: Any | None = ..., stripe_account: Any | None = ..., last_response: Any | None = ...
    ): ...
    api_key: Any
    stripe_version: Any
    stripe_account: Any
    def refresh_from(
        self,
        values,
        api_key: Any | None = ...,
        partial: bool = ...,
        stripe_version: Any | None = ...,
        stripe_account: Any | None = ...,
        last_response: Any | None = ...,
    ) -> None: ...
    @classmethod
    def api_base(cls) -> None: ...
    def request(self, method, url, params: Any | None = ..., headers: Any | None = ...): ...
    def request_stream(self, method, url, params: Any | None = ..., headers: Any | None = ...): ...
    def to_dict(self): ...
    def to_dict_recursive(self): ...
    @property
    def stripe_id(self): ...
    def serialize(self, previous): ...
    def __copy__(self) -> StripeObject: ...
    def __deepcopy__(self, memo: Any) -> StripeObject: ...



from typing import Any

def modify_dn_operation(dn, new_relative_dn, delete_old_rdn: bool = ..., new_superior: Any | None = ...): ...
def modify_dn_request_to_dict(request): ...
def modify_dn_response_to_dict(response): ...



import pickle
import sys
from abc import ABCMeta
from copyreg import _DispatchTableType
from typing import Any
from typing_extensions import Literal

if sys.platform == "win32":
    __all__ = ["send_handle", "recv_handle", "ForkingPickler", "register", "dump", "DupHandle", "duplicate", "steal_handle"]
else:
    __all__ = ["send_handle", "recv_handle", "ForkingPickler", "register", "dump", "DupFd", "sendfds", "recvfds"]

class ForkingPickler(pickle.Pickler):
    dispatch_table: _DispatchTableType
    def __init__(self, *args) -> None: ...
    @classmethod
    def register(cls, type, reduce) -> None: ...
    @classmethod
    def dumps(cls, obj, protocol: Any | None = ...): ...
    loads = pickle.loads

register = ForkingPickler.register

def dump(obj, file, protocol: Any | None = ...) -> None: ...

if sys.platform == "win32":
    if sys.version_info >= (3, 8):
        def duplicate(handle, target_process: Any | None = ..., inheritable: bool = ..., *, source_process: Any | None = ...): ...
    else:
        def duplicate(handle, target_process: Any | None = ..., inheritable: bool = ...): ...

    def steal_handle(source_pid, handle): ...
    def send_handle(conn, handle, destination_pid) -> None: ...
    def recv_handle(conn): ...

    class DupHandle:
        def __init__(self, handle, access, pid: Any | None = ...) -> None: ...
        def detach(self): ...

else:
    if sys.platform == "darwin":
        ACKNOWLEDGE: Literal[True]
    else:
        ACKNOWLEDGE: Literal[False]

    def recvfds(sock, size): ...
    def send_handle(conn, handle, destination_pid) -> None: ...
    def recv_handle(conn) -> None: ...
    def sendfds(sock, fds) -> None: ...
    def DupFd(fd): ...

# These aliases are to work around pyright complaints.
# Pyright doesn't like it when a class object is defined as an alias
# of a global object with the same name.
_ForkingPickler = ForkingPickler
_register = register
_dump = dump
_send_handle = send_handle
_recv_handle = recv_handle

if sys.platform == "win32":
    _steal_handle = steal_handle
    _duplicate = duplicate
    _DupHandle = DupHandle
else:
    _sendfds = sendfds
    _recvfds = recvfds
    _DupFd = DupFd

class AbstractReducer(metaclass=ABCMeta):
    ForkingPickler = _ForkingPickler
    register = _register
    dump = _dump
    send_handle = _send_handle
    recv_handle = _recv_handle
    if sys.platform == "win32":
        steal_handle = _steal_handle
        duplicate = _duplicate
        DupHandle = _DupHandle
    else:
        sendfds = _sendfds
        recvfds = _recvfds
        DupFd = _DupFd
    def __init__(self, *args) -> None: ...



from _typeshed import Self
from collections.abc import Callable, Iterable, Mapping, Sequence
from types import TracebackType
from typing import Any, TypeVar, overload

import psycopg2
import psycopg2.extensions
from psycopg2.sql import Composable

_Vars = Sequence[Any] | Mapping[str, Any] | None

BINARY: Any
BINARYARRAY: Any
BOOLEAN: Any
BOOLEANARRAY: Any
BYTES: Any
BYTESARRAY: Any
CIDRARRAY: Any
DATE: Any
DATEARRAY: Any
DATETIME: Any
DATETIMEARRAY: Any
DATETIMETZ: Any
DATETIMETZARRAY: Any
DECIMAL: Any
DECIMALARRAY: Any
FLOAT: Any
FLOATARRAY: Any
INETARRAY: Any
INTEGER: Any
INTEGERARRAY: Any
INTERVAL: Any
INTERVALARRAY: Any
LONGINTEGER: Any
LONGINTEGERARRAY: Any
MACADDRARRAY: Any
NUMBER: Any
PYDATE: Any
PYDATEARRAY: Any
PYDATETIME: Any
PYDATETIMEARRAY: Any
PYDATETIMETZ: Any
PYDATETIMETZARRAY: Any
PYINTERVAL: Any
PYINTERVALARRAY: Any
PYTIME: Any
PYTIMEARRAY: Any
REPLICATION_LOGICAL: int
REPLICATION_PHYSICAL: int
ROWID: Any
ROWIDARRAY: Any
STRING: Any
STRINGARRAY: Any
TIME: Any
TIMEARRAY: Any
UNICODE: Any
UNICODEARRAY: Any
UNKNOWN: Any
adapters: dict[Any, Any]
apilevel: str
binary_types: dict[Any, Any]
encodings: dict[Any, Any]
paramstyle: str
sqlstate_errors: dict[Any, Any]
string_types: dict[Any, Any]
threadsafety: int

__libpq_version__: int

class AsIs:
    adapted: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def __conform__(self, *args, **kwargs): ...

class Binary:
    adapted: Any
    buffer: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def prepare(self, conn): ...
    def __conform__(self, *args, **kwargs): ...

class Boolean:
    adapted: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def __conform__(self, *args, **kwargs): ...

class Column:
    display_size: Any
    internal_size: Any
    name: Any
    null_ok: Any
    precision: Any
    scale: Any
    table_column: Any
    table_oid: Any
    type_code: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def __eq__(self, __other): ...
    def __ge__(self, __other): ...
    def __getitem__(self, __index): ...
    def __getstate__(self): ...
    def __gt__(self, __other): ...
    def __le__(self, __other): ...
    def __len__(self): ...
    def __lt__(self, __other): ...
    def __ne__(self, __other): ...
    def __setstate__(self, state): ...

class ConnectionInfo:
    backend_pid: Any
    dbname: Any
    dsn_parameters: Any
    error_message: Any
    host: Any
    needs_password: Any
    options: Any
    password: Any
    port: Any
    protocol_version: Any
    server_version: Any
    socket: Any
    ssl_attribute_names: Any
    ssl_in_use: Any
    status: Any
    transaction_status: Any
    used_password: Any
    user: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def parameter_status(self, *args, **kwargs): ...
    def ssl_attribute(self, *args, **kwargs): ...

class DataError(psycopg2.DatabaseError): ...
class DatabaseError(psycopg2.Error): ...

class Decimal:
    adapted: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def __conform__(self, *args, **kwargs): ...

class Diagnostics:
    column_name: Any
    constraint_name: Any
    context: Any
    datatype_name: Any
    internal_position: Any
    internal_query: Any
    message_detail: Any
    message_hint: Any
    message_primary: Any
    schema_name: Any
    severity: Any
    severity_nonlocalized: Any
    source_file: Any
    source_function: Any
    source_line: Any
    sqlstate: Any
    statement_position: Any
    table_name: Any
    def __init__(self, *args, **kwargs) -> None: ...

class Error(Exception):
    cursor: Any
    diag: Any
    pgcode: Any
    pgerror: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def __reduce__(self): ...
    def __setstate__(self, state): ...

class Float:
    adapted: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def __conform__(self, *args, **kwargs): ...

class ISQLQuote:
    _wrapped: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getbinary(self, *args, **kwargs): ...
    def getbuffer(self, *args, **kwargs): ...
    def getquoted(self, *args, **kwargs): ...

class Int:
    adapted: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def __conform__(self, *args, **kwargs): ...

class IntegrityError(psycopg2.DatabaseError): ...
class InterfaceError(psycopg2.Error): ...
class InternalError(psycopg2.DatabaseError): ...

class List:
    adapted: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def prepare(self, *args, **kwargs): ...
    def __conform__(self, *args, **kwargs): ...

class NotSupportedError(psycopg2.DatabaseError): ...

class Notify:
    channel: Any
    payload: Any
    pid: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def __eq__(self, __other): ...
    def __ge__(self, __other): ...
    def __getitem__(self, __index): ...
    def __gt__(self, __other): ...
    def __hash__(self): ...
    def __le__(self, __other): ...
    def __len__(self): ...
    def __lt__(self, __other): ...
    def __ne__(self, __other): ...

class OperationalError(psycopg2.DatabaseError): ...
class ProgrammingError(psycopg2.DatabaseError): ...
class QueryCanceledError(psycopg2.OperationalError): ...

class QuotedString:
    adapted: Any
    buffer: Any
    encoding: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def getquoted(self, *args, **kwargs): ...
    def prepare(self, *args, **kwargs): ...
    def __conform__(self, *args, **kwargs): ...

class ReplicationConnection(psycopg2.extensions.connection):
    autocommit: Any
    isolation_level: Any
    replication_type: Any
    reset: Any
    set_isolation_level: Any
    set_session: Any
    def __init__(self, *args, **kwargs) -> None: ...

class ReplicationCursor(psycopg2.extensions.cursor):
    feedback_timestamp: Any
    io_timestamp: Any
    wal_end: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def consume_stream(self, consumer, keepalive_interval=...): ...
    def read_message(self, *args, **kwargs): ...
    def send_feedback(self, write_lsn=..., flush_lsn=..., apply_lsn=..., reply=..., force=...): ...
    def start_replication_expert(self, command, decode=..., status_interval=...): ...

class ReplicationMessage:
    cursor: Any
    data_size: Any
    data_start: Any
    payload: Any
    send_time: Any
    wal_end: Any
    def __init__(self, *args, **kwargs) -> None: ...

class TransactionRollbackError(psycopg2.OperationalError): ...
class Warning(Exception): ...

class Xid:
    bqual: Any
    database: Any
    format_id: Any
    gtrid: Any
    owner: Any
    prepared: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def from_string(self, *args, **kwargs): ...
    def __getitem__(self, __index): ...
    def __len__(self): ...

_cursor = cursor
_T_cur = TypeVar("_T_cur", bound=_cursor)

class connection:
    DataError: Any
    DatabaseError: Any
    Error: Any
    IntegrityError: Any
    InterfaceError: Any
    InternalError: Any
    NotSupportedError: Any
    OperationalError: Any
    ProgrammingError: Any
    Warning: Any
    async_: Any
    autocommit: Any
    binary_types: Any
    closed: Any
    cursor_factory: Callable[..., _cursor]
    deferrable: Any
    dsn: Any
    encoding: Any
    info: Any
    isolation_level: Any
    notices: Any
    notifies: Any
    pgconn_ptr: Any
    protocol_version: Any
    readonly: Any
    server_version: Any
    status: Any
    string_types: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def cancel(self, *args, **kwargs): ...
    def close(self, *args, **kwargs): ...
    def commit(self, *args, **kwargs): ...
    @overload
    def cursor(self, name=..., *, scrollable=..., withhold=...) -> _cursor: ...
    @overload
    def cursor(self, name=..., cursor_factory: Callable[..., _T_cur] = ..., scrollable=..., withhold=...) -> _T_cur: ...
    def fileno(self, *args, **kwargs): ...
    def get_backend_pid(self, *args, **kwargs): ...
    def get_dsn_parameters(self, *args, **kwargs): ...
    def get_native_connection(self, *args, **kwargs): ...
    def get_parameter_status(self, parameter): ...
    def get_transaction_status(self): ...
    def isexecuting(self, *args, **kwargs): ...
    def lobject(self, oid=..., mode=..., new_oid=..., new_file=..., lobject_factory=...): ...
    def poll(self, *args, **kwargs): ...
    def reset(self): ...
    def rollback(self): ...
    def set_client_encoding(self, encoding): ...
    def set_isolation_level(self, level): ...
    def set_session(self, *args, **kwargs): ...
    def tpc_begin(self, xid): ...
    def tpc_commit(self, *args, **kwargs): ...
    def tpc_prepare(self): ...
    def tpc_recover(self): ...
    def tpc_rollback(self, *args, **kwargs): ...
    def xid(self, format_id, gtrid, bqual): ...
    def __enter__(self): ...
    def __exit__(self, type, value, traceback): ...

class cursor:
    arraysize: int
    binary_types: Any
    closed: Any
    connection: Any
    description: Any
    itersize: Any
    lastrowid: Any
    name: Any
    pgresult_ptr: Any
    query: Any
    row_factory: Any
    rowcount: int
    rownumber: int
    scrollable: Any
    statusmessage: Any
    string_types: Any
    typecaster: Any
    tzinfo_factory: Any
    withhold: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def callproc(self, procname, parameters=...): ...
    def cast(self, oid, s): ...
    def close(self): ...
    def copy_expert(self, sql: str | bytes | Composable, file, size=...): ...
    def copy_from(self, file, table, sep=..., null=..., size=..., columns=...): ...
    def copy_to(self, file, table, sep=..., null=..., columns=...): ...
    def execute(self, query: str | bytes | Composable, vars: _Vars = ...) -> None: ...
    def executemany(self, query: str | bytes | Composable, vars_list: Iterable[_Vars]) -> None: ...
    def fetchall(self) -> list[tuple[Any, ...]]: ...
    def fetchmany(self, size=...) -> list[tuple[Any, ...]]: ...
    def fetchone(self) -> tuple[Any, ...] | None: ...
    def mogrify(self, *args, **kwargs): ...
    def nextset(self): ...
    def scroll(self, value, mode=...): ...
    def setinputsizes(self, sizes): ...
    def setoutputsize(self, size, column=...): ...
    def __enter__(self: Self) -> Self: ...
    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None: ...
    def __iter__(self: Self) -> Self: ...
    def __next__(self) -> tuple[Any, ...]: ...

class lobject:
    closed: Any
    mode: Any
    oid: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def close(self): ...
    def export(self, filename): ...
    def read(self, size=...): ...
    def seek(self, offset, whence=...): ...
    def tell(self): ...
    def truncate(self, len=...): ...
    def unlink(self): ...
    def write(self, str): ...

def Date(year, month, day): ...
def DateFromPy(*args, **kwargs): ...
def DateFromTicks(ticks): ...
def IntervalFromPy(*args, **kwargs): ...
def Time(hour, minutes, seconds, tzinfo=...): ...
def TimeFromPy(*args, **kwargs): ...
def TimeFromTicks(ticks): ...
def Timestamp(year, month, day, hour, minutes, seconds, tzinfo=...): ...
def TimestampFromPy(*args, **kwargs): ...
def TimestampFromTicks(ticks): ...
def _connect(*args, **kwargs): ...
def adapt(*args, **kwargs): ...
def encrypt_password(*args, **kwargs): ...
def get_wait_callback(*args, **kwargs): ...
def libpq_version(*args, **kwargs): ...
def new_array_type(oids, name, baseobj): ...
def new_type(oids, name, castobj): ...
def parse_dsn(*args, **kwargs): ...
def quote_ident(*args, **kwargs): ...
def register_type(*args, **kwargs): ...
def set_wait_callback(_none): ...



from array import array
from mmap import mmap
from typing import Any, Text

class error(Exception): ...

_FmtType = bytes | Text
_BufferType = array[int] | bytes | bytearray | buffer | memoryview | mmap
_WriteBufferType = array[Any] | bytearray | buffer | memoryview | mmap

def pack(fmt: _FmtType, *v: Any) -> bytes: ...
def pack_into(fmt: _FmtType, buffer: _WriteBufferType, offset: int, *v: Any) -> None: ...
def unpack(__format: _FmtType, __buffer: _BufferType) -> tuple[Any, ...]: ...
def unpack_from(__format: _FmtType, buffer: _BufferType, offset: int = ...) -> tuple[Any, ...]: ...
def calcsize(__format: _FmtType) -> int: ...

class Struct:
    format: bytes
    size: int
    def __init__(self, format: _FmtType) -> None: ...
    def pack(self, *v: Any) -> bytes: ...
    def pack_into(self, buffer: _WriteBufferType, offset: int, *v: Any) -> None: ...
    def unpack(self, __buffer: _BufferType) -> tuple[Any, ...]: ...
    def unpack_from(self, buffer: _BufferType, offset: int = ...) -> tuple[Any, ...]: ...
